
__copyright__ = 'Copyright 2013-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import time

import threading          as mt

import radical.utils      as ru

from ... import states    as rps
from ... import agent     as rpa
from ... import constants as rpc
from ... import utils     as rpu


# ------------------------------------------------------------------------------
# 'enum' for RP's spawner types
EXECUTING_NAME_POPEN   = 'POPEN'
EXECUTING_NAME_FLUX    = 'FLUX'
EXECUTING_NAME_SLEEP   = 'SLEEP'


# ------------------------------------------------------------------------------
#
class AgentExecutingComponent(rpu.Component):
    '''
    Manage the creation of Task processes, and watch them until they are
    completed (one way or the other).  The spawner thus moves the task from
    PendingExecution to Executing, and then to a final state (or PendingStageOut
    of course).
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        rpu.Component.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Spawner
    #
    @classmethod
    def create(cls, cfg, session):

        # Make sure that we are the base-class!
        if cls != AgentExecutingComponent:
            raise TypeError('Factory only available to base class!')

        name = cfg['spawner']

        from .popen    import Popen
        from .flux     import Flux
        from .sleep    import Sleep

        impl = {
            EXECUTING_NAME_POPEN  : Popen,
            EXECUTING_NAME_FLUX   : Flux,
            EXECUTING_NAME_SLEEP  : Sleep,
        }

        if name not in impl:
            raise ValueError('AgentExecutingComponent %s unknown' % name)

        return impl[name](cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

      # self._log.debug('exec base initialize')

        # The spawner/executor needs the ResourceManager information which have
        # been collected during agent startup.
        self._rm = rpa.ResourceManager.create(self._cfg.resource_manager,
                                              self._cfg, self._log, self._prof)

        self._pwd      = os.path.realpath(os.getcwd())
        self.sid       = self._cfg['sid']
        self.resource  = self._cfg['resource']
        self.rsbox     = self._cfg['resource_sandbox']
        self.ssbox     = self._cfg['session_sandbox']
        self.psbox     = self._cfg['pilot_sandbox']
        self.gtod      = '$RP_PILOT_SANDBOX/gtod'
        self.prof      = '$RP_PILOT_SANDBOX/prof'

        # if so configured, let the Task know what to use as tmp dir
        self._task_tmp = self._cfg.get('task_tmp',
                                       os.environ.get('TMP', '/tmp'))


        if self.psbox.startswith(self.ssbox):
            self.psbox = '$RP_SESSION_SANDBOX%s'  % self.psbox[len(self.ssbox):]
        if self.ssbox.startswith(self.rsbox):
            self.ssbox = '$RP_RESOURCE_SANDBOX%s' % self.ssbox[len(self.rsbox):]
        if self.ssbox.endswith(self.sid):
            self.ssbox = '%s$RP_SESSION_ID/'      % self.ssbox[:-len(self.sid)]

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self.control_cb)

        self._to_tasks  = list()
        self._to_lock   = mt.Lock()
        self._to_thread = mt.Thread(target=self._to_watcher)
        self._to_thread.daemon = True
        self._to_thread.start()


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        raise NotImplementedError('work is not implemented')


    # --------------------------------------------------------------------------
    #
    def control_cb(self, topic, msg):

        self._log.info('command_cb [%s]: %s', topic, msg)

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_tasks':

            self._log.info('cancel_tasks command (%s)', arg)
            for tid in arg['uids']:
                self.cancel_task(tid)

        return True


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, uid):

        raise NotImplementedError('cancel_task is not implemented')


    # --------------------------------------------------------------------------
    #
    def _to_watcher(self):
        '''
        watch the set of tasks for which timeouts are defined.  If the timeout
        passes and the tasks are still active, kill the task via
        `self._cancel_task(task)`.  That has to be implemented by al executors.
        '''

        while True:

            # check once per second at most
            time.sleep(1)

            now = time.time()
            with self._to_lock:

                # running tasks for next check
                to_list = list()
                for to, start, task in self._to_tasks:
                    if now - start > to:
                        self._prof.prof('task_timeout', uid=task['uid'])
                        self.cancel_task(uid=task['uid'])
                    else:
                        to_list.append([to, start, task])

                self._to_tasks = to_list


    # --------------------------------------------------------------------------
    #
    def handle_timeout(self, task):

        to = task['description'].get('timeout', 0.0)

        if to > 0.0:
            with self._to_lock:
                self._to_tasks.append([to, time.time(), task])


    # --------------------------------------------------------------------------
    #
    def advance_tasks(self, tasks, state, publish, push, ts=None):
        '''
        sort tasks into different buckets, depending on their origin.
        That origin will determine where tasks which completed execution
        and end up here will be routed to:

          - client: state update to update worker
          - raptor: state update to `STATE_PUBSUB`
          - agent : state update to `STATE_PUBSUB`

        a fallback is not in place to enforce the specification of the
        `origin` attributes for tasks.
        '''


        buckets = {'client': list(),
                   'raptor': list(),
                   'agent' : list()}

        for task in ru.as_list(tasks):
            buckets[task['origin']].append(task)

        # we want any task which has a `raptor_id` set to show up in raptor's
        # result callbacks
        if state != rps.AGENT_EXECUTING:
            for task in ru.as_list(tasks):
                if task['description'].get('raptor_id'):
                    if task not in buckets['raptor']:
                        buckets['raptor'].append(task)

        if buckets['client']:
            self.advance(buckets['client'], state=state,
                                            publish=publish, push=push, ts=ts)

        if buckets['raptor']:
            self.advance(buckets['raptor'], state=state,
                                            publish=publish, push=False, ts=ts)
            self.publish(rpc.STATE_PUBSUB, {'cmd': 'raptor_state_update',
                                            'arg': buckets['raptor']})

        if buckets['agent']:
            self.advance(buckets['agent'], state=state,
                                            publish=publish, push=False, ts=ts)


# ------------------------------------------------------------------------------





from typing import Dict, List, Any, Optional

import radical.utils as ru

from .task             import Task
from .task_description import TaskDescription


# ------------------------------------------------------------------------------
#
class Raptor(Task):
    '''
    RAPTOR ('RAPid Task executOR') is a task executor which, other than other
    RADICAL-Pilot executors can handle function tasks.

    A `Raptor` must be submitted to a pilot.  It will be associated with
    `RaptorWorker` instances on that pilot and use those workers to rapidly
    execute tasks.  Raptors excel at high throughput execution for large numbers
    of short running tasks.  However, they have limited capabilities with
    respect to managing task data dependencies, multinode tasks, MPI
    executables, and tasks with heterogeneous resource requirements.
    '''

    # --------------------------------------------------------------------------
    #
    def submit_workers(self, descriptions: List[TaskDescription]) -> List[Task]:
        '''
        Submit a set of workers for this `Raptor` instance.

        Args:
            descriptions (List[TaskDescription]): ;aunch a raptor worker for
                each provided description.

        Returns:
            List[Tasks]: a list of `rp.Task` instances, one for each created
                worker task

        The method will return immediately without waiting for actual task
        instantiation.  The submitted tasks will operate solely on behalf of the
        `Raptor` master this method is being called on.
        '''

        descriptions = ru.as_list(descriptions)

        for td in descriptions:
            td.raptor_id = self.uid

        return self._tmgr.submit_workers(descriptions)


    # --------------------------------------------------------------------------
    #
    def submit_tasks(self, descriptions: List[TaskDescription]) -> List[Task]:
        '''
        Submit a set of tasks to this `Raptor` instance.

        Args:
            descriptions (List[TaskDescription]): ;aunch a raptor worker for
                each provided description.

        Returns:
            List[Tasks]: a list of `rp.Task` instances, one for each task.

        The tasks might not get executed until a worker is available for this
        Raptor instance.
        '''



        descriptions = ru.as_list(descriptions)

        for td in descriptions:
            td.raptor_id = self.uid

        return self._tmgr.submit_tasks(descriptions)


    # --------------------------------------------------------------------------
    #
    def rpc(self, rpc: str,
                  args: Optional[Dict[str, Any]] = None
           ) -> Dict[str, Any]:
        '''
        Send a raptor command, wait for the response, and return the result.

        Args:
            rpc (str): name of the rpc call to invoke
            args (Dict[str, Any])): kwargs dict for the rpc call

        Returns:
            Dict[str, Any]: the returned dictionary has the following fields:
              - out: captured standard output
              - err: captured standard error
              - ret: return value of the call (can be any serializable type)
              - exc: tuple of exception type (str) and error message (str)
        '''

        if not self._pilot:
            raise RuntimeError('not assoigned to a pilot yet, cannot run rpc')

        reply = self._session._dbs.pilot_rpc(self._pilot, self.uid, rpc, args)

        return reply


# ------------------------------------------------------------------------------
#
RaptorMaster = Raptor
'''
The `RaptorMaster` task is an alias for the `Raptor` class and is provided
for naming symmetry with `RaptorWorker`.
'''


# ------------------------------------------------------------------------------
#
class RaptorWorker(Task):
    '''
    A `RaptorWorker` task is a thin wrapper around a normal `rp.Task` instance
    for the purpose of representing a Raptor worker task.

    NOTE: this class is not yet used.
    '''

    pass


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'


# configure the psij logger (captured in the launch components stderr)
import logging
logging.basicConfig(level='DEBUG')

import threading as mt

from .base import PilotLauncherBase
from ...   import states as rps

# psij is optional
psij    = None
psij_ex = None

try:
    import psij
except ImportError as ex:
    psij_ex = ex


# ------------------------------------------------------------------------------
#
class PilotLauncherPSIJ(PilotLauncherBase):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, log, prof, state_cb):


        # psij is an optional dependency - let an import exception fall through
        # to disable this pilot launcher
        if psij_ex:
            raise psij_ex

        assert psij

        PilotLauncherBase.__init__(self, name, log, prof, state_cb)

        self._jobs   = dict()      # map pilot id to psi_j job
        self._pilots = dict()      # map psi_j id to pilot job
        self._jex    = dict()      # map launch schema to psi_j job executors
        self._lock   = mt.RLock()  # lock above structures


    # --------------------------------------------------------------------------
    #
    def _get_schema(self, rcfg):

        url     = rcfg['job_manager_endpoint']
        schemas = url.split(':')[0].split('+')

        if 'ssh' in schemas:
            schemas.remove('ssh')

        if 'gsissh' in schemas:
            schemas.remove('gsissh')

        if len(schemas) > 1:
            return

        if not schemas:
            return

        schema = schemas[0]

        if schema == 'fork':
            schema = 'local'

        return schema


    # --------------------------------------------------------------------------
    #
    def _translate_state(self, status):

        if   status.state == psij.JobState.NEW       : return rps.NEW
        elif status.state == psij.JobState.QUEUED    : return rps.PMGR_LAUNCHING
        elif status.state == psij.JobState.ACTIVE    : return rps.PMGR_ACTIVE
        elif status.state == psij.JobState.COMPLETED : return rps.DONE
        elif status.state == psij.JobState.FAILED    : return rps.FAILED
        elif status.state == psij.JobState.CANCELED  : return rps.CANCELED
        else:
            raise ValueError('cannot interpret psij state: %s' % repr(status))


    # --------------------------------------------------------------------------
    #
    def _job_status_cb(self, job, status):

        try:
            with self._lock:

                if job.id not in self._pilots:
                    return

                rp_state = self._translate_state(status)
                pilot    = self._pilots[job.id]

            self._state_cb(pilot, rp_state)

        except Exception:
            self._log.exception('job status callback failed')


    # --------------------------------------------------------------------------
    #
    def can_launch(self, rcfg, pilot):

        schema = self._get_schema(rcfg)

        if not schema:
            return False

        if schema not in self._jex:

            self._log.debug('create executor for %s', schema)
            try:
                self._jex[schema] = psij.JobExecutor.get_instance(schema)
                self._jex[schema].set_job_status_callback(self._job_status_cb)
            except:
                self._log.exception('failed to create psij executor')
                return False

        return True


    # --------------------------------------------------------------------------
    #
    def launch_pilots(self, rcfg, pilots):

        assert psij

        for pilot in pilots:

            pid    = pilot['uid']
            schema = self._get_schema(rcfg)
            assert schema

            jex = self._jex.get(schema)

            assert jex

            jd = pilot['jd_dict']

            proj, res = None, None
            if jd.project:
                if ':' in jd.project:
                    proj, res = jd.project.split(':', 1)
                else:
                    proj = jd.project

            attr = psij.JobAttributes()
            attr.duration       = jd.wall_time_limit
            attr.queue_name     = jd.queue
            attr.project_name   = proj
            attr.reservation_id = res

            spec = psij.JobSpec()
            spec.attributes  = attr
            spec.executable  = jd.executable
            spec.arguments   = jd.arguments
            spec.environment = jd.environment
            spec.directory   = jd.working_directory
            spec.stdout_path = jd.output
            spec.stderr_path = jd.error

            spec.resources   = psij.ResourceSpecV1()
            spec.resources.node_count            = jd.node_count
            spec.resources.process_count         = jd.total_cpu_count
          # spec.resources.cpu_cores_per_process = 1
          # spec.resources.gpu_cores_per_process = jd.total_gpu_count

            job = psij.Job(spec)

            self._jobs[pid]      = job
            self._pilots[job.id] = pilot
            self._log.debug('added %s: %s', job.id, pid)

            jex.submit(job)


    # --------------------------------------------------------------------------
    #
    def kill_pilots(self, pids):

        for pid in pids:
            if pid not in pids:
                continue

            self._jobs[pid].cancel()


# ------------------------------------------------------------------------------




# pylint: disable=protected-access

__copyright__ = 'Copyright 2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import functools

import threading as mt

# saga is optional
rs    = None
rs_ex = None

try:
    import radical.saga
    rs = radical.saga
except ImportError as ex:
    rs_ex = ex

import radical.utils as ru

from .base import PilotLauncherBase
from ...   import states as rps


# ------------------------------------------------------------------------------
#
class PilotLauncherSAGA(PilotLauncherBase):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, log, prof, state_cb):

        if rs_ex:
            raise rs_ex

        assert rs

        PilotLauncherBase.__init__(self, name, log, prof, state_cb)

        self._jobs   = dict()      # pid      : rs.Job
        self._js     = dict()      # resource : rs.JobService
        self._pilots = dict()      # saga_id  : pilot job
        self._lock   = mt.RLock()  # lock for above


        # FIXME: get session from launching component
        self._session = rs.Session()


    # --------------------------------------------------------------------------
    #
    def _translate_state(self, saga_state):

        if   saga_state == rs.job.NEW       : return rps.NEW
        elif saga_state == rs.job.PENDING   : return rps.PMGR_LAUNCHING
        elif saga_state == rs.job.RUNNING   : return rps.PMGR_LAUNCHING
        elif saga_state == rs.job.SUSPENDED : return rps.PMGR_LAUNCHING
        elif saga_state == rs.job.DONE      : return rps.DONE
        elif saga_state == rs.job.FAILED    : return rps.FAILED
        elif saga_state == rs.job.CANCELED  : return rps.CANCELED
        else:
            raise ValueError('cannot interpret psij state: %s' % saga_state)


    # --------------------------------------------------------------------------
    #
    def _job_state_cb(self, job, _, saga_state, pid):

        try:
            with self._lock:

                if pid not in self._pilots:
                    return

                rp_state = self._translate_state(saga_state)
                pilot    = self._pilots[pid]

            self._state_cb(pilot, rp_state)

        except Exception:
            self._log.exception('job status callback failed')

        return True


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        # FIXME: terminate thread

        with self._lock:

            # cancel pilots
            for job in self._jobs.values():
                job.cancel()

            # close job services
            for url, js in self._js.items():
                self._log.debug('close js %s', url)
                js.close()


    # --------------------------------------------------------------------------
    #
    def can_launch(self, rcfg, pilot):

        # SAGA can launch all pilots
        return True


    # --------------------------------------------------------------------------
    #
    def launch_pilots(self, rcfg, pilots):

        js_ep  = rcfg['job_manager_endpoint']
        self._log.debug('js_ep: %s', js_ep)
        with self._lock:
            if js_ep in self._js:
                js = self._js[js_ep]
            else:
                js = rs.job.Service(js_ep, session=self._session)
                self._js[js_ep] = js

        # now that the scripts are in place and configured,
        # we can launch the agent
        jc = rs.job.Container()

        for pilot in pilots:

            jd_dict = pilot['jd_dict']

            # we need to quote arguments for RS to handle
            jd_dict['arguments'] = [ru.sh_quote(arg)
                                       for arg in jd_dict['arguments']]

            saga_jd_supplement = dict()
            if 'saga_jd_supplement' in jd_dict:
                saga_jd_supplement = jd_dict['saga_jd_supplement']
                del jd_dict['saga_jd_supplement']

            # saga will take care of node_count itself
            if 'node_count' in jd_dict:
                del jd_dict['node_count']

            jd = rs.job.Description()
            for k, v in jd_dict.items():
                jd.set_attribute(k, v)

            # set saga_jd_supplement if not already set
            for key, val in saga_jd_supplement.items():
                if not jd[key]:
                    self._log.debug('supplement %s: %s', key, val)
                    jd[key] = val

            # remember the pilot
            pid = pilot['uid']
            self._pilots[pid] = pilot

            job = js.create_job(jd)
            cb  = functools.partial(self._job_state_cb, pid=pid)
            job.add_callback(rs.STATE, cb)
            jc.add(job)

        jc.run()

        # Order of tasks in `rs.job.Container().tasks` is not changing over the
        # time, thus it's able to iterate over it and other list(s) all together
        for j, pilot in zip(jc.get_tasks(), pilots):

            # do a quick error check
            if j.state == rs.FAILED:
                self._log.error('%s: %s : %s : %s', j.id, j.state, j.stderr, j.stdout)
                raise RuntimeError("SAGA Job state is FAILED. (%s)" % j.name)

            pid = pilot['uid']

            # FIXME: update the right pilot
            with self._lock:
                self._jobs[pid] = j


    # --------------------------------------------------------------------------
    #
    def kill_pilots(self, pids):

        tc = rs.job.Container()

        for pid in pids:

            job   = self._jobs[pid]
            pilot = self._pilots[pid]

            # don't overwrite resource_details from the agent
            if 'resource_details' in pilot:
                del pilot['resource_details']

            if pilot['state'] in rps.FINAL:
                continue

            self._log.debug('plan cancellation of %s : %s', pilot, job)
            self._log.debug('request cancel for %s', pilot['uid'])
            tc.add(job)

        self._log.debug('cancellation start')
        tc.cancel()
        tc.wait()
        self._log.debug('cancellation done')


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2016-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'


import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
# ibrun: TACC's wrapper around system specific launchers just as srun, mpirun...
#
class IBRun(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        self._command: str  = ''

        LaunchMethod.__init__(self, name, lm_cfg, rm_info, log, prof)


    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):

        lm_info = {'env'    : env,
                   'env_sh' : env_sh,
                   'command': ru.which('ibrun')}

        return lm_info


    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        self._env     = lm_info['env']
        self._env_sh  = lm_info['env_sh']
        self._command = lm_info['command']

        assert self._command


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        pass


    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        if not task['description']['executable']:
            return False, 'no executable'

        return True, ''


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        return ['. $RP_PILOT_SANDBOX/%s' % self._env_sh]


    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        slots = task['slots']
        assert slots.get('ranks'), 'task.slots.ranks is not set or empty'

        td                 = task['description']
        n_ranks            = td['ranks']
        n_threads_per_rank = td['cores_per_rank']

        # Usage of env variable IBRUN_TASKS_PER_NODE is purely for MPI tasks,
        # threads are not considered (info provided by TACC support)
        tasks_per_node = self._lm_cfg.get('options', {}).get('tasks_per_node')
        if not tasks_per_node:
            tasks_per_node = self._rm_info['cores_per_node'] // \
                             (n_ranks * n_threads_per_rank) or 1
        # IBRUN_TASKS_PER_NODE is used to set the actual number of running
        # tasks, if not set, then `ibrun` script will use the default Slurm
        # setting for the number of tasks per node to build the hostlist
        # (for TACC machines)
        #
        # NOTE: in case of performance issue: reconsider this parameter
        launch_env = 'IBRUN_TASKS_PER_NODE=%d' % tasks_per_node

        rank_node_ids = set([n['node_id'] for n in slots['ranks']])
        tasks_offset  = 0
        ibrun_offset  = 0

        for node in self._rm_info.node_list:

            if node['node_id'] not in rank_node_ids:
                tasks_offset += tasks_per_node
                continue

            # core_map contains core ids for each thread,
            # but threads are ignored for offset
            core_id_min = min([rank['core_map'][0][0]
                               for rank in slots['ranks']
                               if rank['node_id'] == node['node_id']])
            # offset into processor (cpus) hostlist
            ibrun_offset = tasks_offset + (core_id_min // n_threads_per_rank)
            break

        cmd = '%s %s -n %s -o %d %s' % (launch_env, self._command, n_ranks,
                                        ibrun_offset, exec_path)
        return cmd.rstrip()


    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        ret  = 'test -z "$SLURM_PROCID" || export RP_RANK=$SLURM_PROCID\n'
        ret += 'test -z "$MPI_RANK"     || export RP_RANK=$MPI_RANK\n'
        ret += 'test -z "$PMIX_RANK"    || export RP_RANK=$PMIX_RANK\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def get_exec(self, task):

        td          = task['description']
        task_exec   = td['executable']
        task_args   = td.get('arguments')
        task_argstr = self._create_arg_string(task_args)
        command     = '%s %s' % (task_exec, task_argstr)

        return command.rstrip()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import sys
import os
import radical.pilot as rp


""" DESCRIPTION: Tutorial 4: A workload consisting of of MPI tasks
"""

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


# ------------------------------------------------------------------------------
#
def pilot_state_cb (pilot, state):

    if not pilot:
        return

    print("[Callback]: Pilot '%s' state: %s." % (pilot.uid, state))

    if state == rp.FAILED:
        sys.exit (1)


# ------------------------------------------------------------------------------
#
def task_state_cb (task, state):

    if not task:
        return

    global CNT

    print("[Callback]: task %s on %s: %s." % (task.uid, task.pilot_id, state))

    if state == rp.FAILED:
        print("stderr: %s" % task.stderr)
        sys.exit(2)


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # we can optionally pass session name to RP
    if len(sys.argv) > 1:
        session_name = sys.argv[1]
    else:
        session_name = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session(name=session_name,
                         dburl=os.environ.get('RADICAL_PILOT_DBURL'))
    print("session id: %s" % session.uid)

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        #
        # Change the user name below if you are using a remote resource
        # and your username on that resource is different from the username
        # on your local machine.
        #
        c = rp.Context('userpass')
      # c.user_id = "tutorial_X"
      # c.user_pass = "PutYourPasswordHere"
        session.add_context(c)

        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        print("Initializing Pilot Manager ...")
        pmgr = rp.PilotManager(session=session)

        # Register our callback with the PilotManager. This callback will get
        # called every time any of the pilots managed by the PilotManager
        # change their state.
        pmgr.register_callback(pilot_state_cb)

        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        #
        # Change the resource below if you want to run on a another resource than
        # the pre-configured tutorial cluster.
        # You also might have to set the 'project' to your allocation ID if
        # your remote resource requires so.
        #
        # A list of pre-configured resources can be found at:
        # https://radicalpilot.readthedocs.io/en/stable/ \
        #        machconf.html#preconfigured-resources
        #
        pdesc = rp.PilotDescription ()
        pdesc.resource = "local.localhost"
        pdesc.runtime  = 10
        pdesc.cores    = 16
        pdesc.cleanup  = True

        # submit the pilot.
        print("Submitting  Pilot to Pilot Manager ...")
        pilot = pmgr.submit_pilots(pdesc)

        # Combine the Pilot, the Tasks and a scheduler via
        # a TaskManager object.
        print("Initializing Task Manager ...")
        tmgr = rp.TaskManager (session=session,
                               scheduler=rp.SCHEDULER_DIRECT_SUBMISSION)

        # Register our callback with the TaskManager. This callback will get
        # called every time any of the tasks managed by the TaskManager
        # change their state.
        tmgr.register_callback(task_state_cb)

        # Add the created Pilot to the TaskManager.
        print("Registering  Pilot with Task Manager ...")
        tmgr.add_pilots(pilot)

        NUMBER_JOBS  = 10  # the total number of tasks to run

        # submit tasks to pilot job
        taskdesc_list = []
        for i in range(NUMBER_JOBS):

            # -------- BEGIN USER DEFINED Task DESCRIPTION --------- #
            taskdesc = rp.TaskDescription()
            taskdesc.executable    = "python"
            taskdesc.arguments     = ["helloworld_mpi.py"]
            taskdesc.input_staging = ["../helloworld_mpi.py"]
            taskdesc.cores         = 8
            taskdesc.mpi           = True
            # -------- END USER DEFINED Task DESCRIPTION --------- #

            taskdesc_list.append(taskdesc)

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        print("Submit Tasks to Task Manager ...")
        task_set = tmgr.submit_tasks (taskdesc_list)

        print("Waiting for tasks to complete ...")
        tmgr.wait_tasks()
        print("All tasks completed successfully!")

        for task in task_set:
            print('* Task %s - state: %s, exit code: %s, started: %s, '
                  'finished: %s, stdout: %s' % (task.uid, task.state,
                  task.exit_code, task.start_time, task.stop_time, task.stdout))

    except Exception as e:
        # Something unexpected happened in the pilot code above
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print("closing session")
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


# ------------------------------------------------------------------------------




# pylint: disable=unused-argument

__copyright__ = 'Copyright 2013-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import stat
import time
import queue
import atexit
import signal
import threading  as mt
import subprocess as sp

import radical.utils as ru

from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentExecutingComponent

from ...task_description import RAPTOR_MASTER, RAPTOR_WORKER


# ------------------------------------------------------------------------------
# ensure tasks are killed on termination
_pids = list()


# pylint: disable=unused-argument
def _kill(*args, **kwargs):

    for pid in _pids:

        # skip test mocks
        if not isinstance(pid, int):
            continue

        try   : os.killpg(pid, signal.SIGTERM)
        except: pass


atexit.register(_kill)
signal.signal(signal.SIGTERM, _kill)
signal.signal(signal.SIGINT,  _kill)
# ------------------------------------------------------------------------------


# ------------------------------------------------------------------------------
#
class Popen(AgentExecutingComponent):

    _header    = '#!/bin/sh\n'
    _separator = '\n# ' + '-' * 78 + '\n'

    # flags for the watcher queue
    TO_WATCH  = 0
    TO_CANCEL = 1

    # --------------------------------------------------------------------------
    #
    def initialize(self):

      # self._log.debug('popen initialize start')
        AgentExecutingComponent.initialize(self)

        self._watch_queue = queue.Queue()

        self._pid = self._cfg['pid']

        # run watcher thread
        self._watcher = mt.Thread(target=self._watch)
      # self._watcher.daemon = True
        self._watcher.start()

      # self._log.debug('popen initialize stop')


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, uid):

        self._watch_queue.put([self.TO_CANCEL, uid])


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        self.advance_tasks(tasks, rps.AGENT_EXECUTING, publish=True, push=False)

        for task in tasks:

            try:
                self._prof.prof('task_start', uid=task['uid'])
                self._handle_task(task)

            except Exception as e:
                self._log.exception("error running Task")
                task['exception']        = repr(e)
                task['exception_detail'] = '\n'.join(ru.get_exception_trace())

                # can't rely on the executor base to free the task resources
                self._prof.prof('unschedule_start', uid=task['uid'])
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

                task['control'] = 'tmgr_pending'
                task['$all']    = True
                self.advance_tasks(task, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def _handle_task(self, task):

        # before we start handling the task, check if it should run in a named
        # env.  If so, inject the activation of that env in the task's pre_exec
        # directives.
        tid  = task['uid']
        td   = task['description']
        sbox = task['task_sandbox_path']

        # prepare stdout/stderr
        task['stdout'] = ''
        task['stderr'] = ''

        stdout_file    = td.get('stdout') or '%s.out' % tid
        stderr_file    = td.get('stderr') or '%s.err' % tid

        if stdout_file[0] != '/':
            task['stdout_file']       = '%s/%s' % (sbox, stdout_file)
            task['stdout_file_short'] = '$RP_TASK_SANDBOX/%s' % stdout_file
        else:
            task['stdout_file']       = stdout_file
            task['stdout_file_short'] = stdout_file

        if stderr_file[0] != '/':
            task['stderr_file']       = '%s/%s' % (sbox, stderr_file)
            task['stderr_file_short'] = '$RP_TASK_SANDBOX/%s' % stderr_file
        else:
            task['stderr_file']       = stderr_file
            task['stderr_file_short'] = stderr_file

        # create two shell scripts: a launcher script (task.launch.sh) which
        # sets the launcher environment, performs pre_launch commands, and then
        # launches the second script which executes the task.
        #
        # The second script (`task.exec.sh`) is instantiated once per task rank.
        # It first resets the environment created by the launcher, then prepares
        # the environment for the tasks.  Next it runs the `pre_exec` directives
        # for all ranks and per rank (if commands for particular rank are
        # defined) are executed, and then, after all ranks are synchronized,
        # finally the task ranks begin to run.
        #
        # The scripts thus show the following approximate structure:
        #
        # Launcher Script (`task.000000.launch.sh`):
        # ----------------------------------------------------------------------
        # #!/bin/sh
        #
        # # `pre_launch` commands
        # date > data.input
        #
        # # launcher specific environment setup
        # module load openmpi
        #
        # # launch the task script
        # mpirun -n 4 ./task.000000.exec.sh
        # ----------------------------------------------------------------------
        #
        # Task Script (`task.000000.exec.sh`)
        # ----------------------------------------------------------------------
        # #!/bin/sh
        #
        # # clean launch environment
        # module unload mpi
        #
        # # task environment setup (`pre_exec`)
        # module load gromacs
        #
        # # rank specific setup
        # touch task.000000.ranks
        # if test "$MPI_RANK" = 0; then
        #   export CUDA_VISIBLE_DEVICES=0
        #   export OMP_NUM_THREADS=2
        #   export RANK_0_VAR=foo
        #   echo 0 >> task.000000.ranks
        # fi
        #
        # if test "$MPI_RANK" = 1; then
        #   export CUDA_VISIBLE_DEVICES=1
        #   export OMP_NUM_THREADS=4
        #   export RANK_1_VAR=bar
        #   echo 1 >> task.000000.ranks
        # fi
        #
        # # synchronize ranks
        # while $(cat task.000000.ranks | wc -l) != $MPI_RANKS; do
        #   sleep 1
        # done
        #
        # # run the task
        # mdrun -i ... -o ... -foo ... 1> task.000000.$MPI_RANK.out \
        #                              2> task.000000.$MPI_RANK.err
        #
        # # now do the very same stuff for the `post_exec` directive
        # ...
        #
        # ----------------------------------------------------------------------
        #
        # NOTE: MongoDB only accepts string keys, and thus the rank IDs in
        #       pre_exec and post_exec dictionaries are rendered as strings.
        #       This should be changed to more intuitive integers once MongoDB
        #       is phased out.
        #

        launcher = self._rm.find_launcher(task)

        if not launcher:
            raise RuntimeError('no launcher found for task %s' % tid)

        self._log.debug('Launching task with %s', launcher.name)

        launch_script = '%s.launch.sh'        % tid
        exec_script   = '%s.exec.sh'          % tid
        exec_path     = '$RP_TASK_SANDBOX/%s' % exec_script

        ru.rec_makedir(sbox)

        if td['mode'] in [RAPTOR_MASTER, RAPTOR_WORKER]:
            ru.write_json('%s/%s.json' % (sbox, tid), td)

        with ru.ru_open('%s/%s' % (sbox, launch_script), 'w') as fout:

            tmp  = ''
            tmp += self._header
            tmp += self._separator
            tmp += self._get_rp_env(task)
            tmp += self._get_rp_funcs()
            tmp += self._separator
            tmp += self._get_prof('launch_start', tid)

            tmp += self._separator
            tmp += '# change to task sandbox\n'
            tmp += 'cd $RP_TASK_SANDBOX\n'

            tmp += self._separator
            tmp += '# prepare launcher env\n'
            tmp += self._get_launch_env(launcher)

            tmp += self._separator
            tmp += '# pre-launch commands\n'
            tmp += self._get_prof('launch_pre', tid)
            tmp += self._get_prep_launch(task, sig='pre_launch')

            tmp += self._separator
            tmp += '# launch commands\n'
            tmp += self._get_prof('launch_submit', tid)
            tmp += self._get_launch(task, launcher, exec_path)
            tmp += self._get_prof('launch_collect', tid)

            tmp += self._separator
            tmp += '# post-launch commands\n'
            tmp += self._get_prof('launch_post', tid)
            tmp += self._get_prep_launch(task, sig='post_launch')

            tmp += self._separator
            tmp += self._get_prof('launch_stop', tid)
            tmp += 'exit $RP_RET\n'

            tmp += self._separator
            tmp += '\n'

            fout.write(tmp)

        # the exec shell script runs the same set of commands for all ranks.
        # However, if the ranks need different GPU's assigned, or if either pre-
        # or post-exec directives contain per-rank dictionaries, then we switch
        # per-rank in the script for all sections between pre- and post-exec.

        n_ranks = td['ranks']
        slots   = task.setdefault('slots', {})

        self._extend_pre_exec(td, slots.get('ranks'))

        with ru.ru_open('%s/%s' % (sbox, exec_script), 'w') as fout:

            tmp  = ''
            tmp += self._header
            tmp += self._separator
            tmp += self._get_rp_env(task)
            tmp += self._get_rp_funcs()
            tmp += self._separator
            tmp += '# rank ID\n'
            tmp += self._get_rank_ids(n_ranks, launcher)
            tmp += self._separator
            tmp += self._get_prof('exec_start', tid)

            tmp += self._get_task_env(task, launcher)

            tmp += self._separator
            tmp += '# pre-exec commands\n'
            tmp += self._get_prof('exec_pre', tid)
            tmp += self._get_prep_exec(task, n_ranks, sig='pre_exec')

            tmp += self._separator
            tmp += '# execute rank\n'
            tmp += self._get_prof('rank_start', tid)
            tmp += self._get_exec(task, launcher)
            tmp += self._get_prof('rank_stop', tid)

            tmp += self._separator
            tmp += '# post-exec commands\n'
            tmp += self._get_prof('exec_post', tid)
            tmp += self._get_prep_exec(task, n_ranks, sig='post_exec')

            tmp += self._separator
            tmp += self._get_prof('exec_stop', tid)
            tmp += 'exit $RP_RET\n'

            tmp += self._separator
            tmp += '\n'

            fout.write(tmp)


        # make sure scripts are executable
        st_l = os.stat('%s/%s' % (sbox, launch_script))
        st_e = os.stat('%s/%s' % (sbox, exec_script))
        os.chmod('%s/%s' % (sbox, launch_script), st_l.st_mode | stat.S_IEXEC)
        os.chmod('%s/%s' % (sbox, exec_script),   st_e.st_mode | stat.S_IEXEC)

        # make sure the sandbox exists
        self._prof.prof('task_mkdir', uid=tid)
        ru.rec_makedir(sbox)
        self._prof.prof('task_mkdir_done', uid=tid)

        # need to set `DEBUG_5` or higher to get slot debug logs
        if self._log._debug_level >= 5:
            ru.write_json('%s/%s.sl' % (sbox, tid), slots)

        # launch and exec script are done, get ready for execution.
        cmdline = '%s/%s' % (sbox, launch_script)

        self._log.info('Launching task %s via %s in %s', tid, cmdline, sbox)

        _launch_out_h = ru.ru_open('%s/%s.launch.out' % (sbox, tid), 'w')
        # `start_new_session=True` is default, which enables decoupling
        # from the parent process group (part of the task cancellation)
        _start_new_session = self._cfg['resource_cfg'].\
            get('new_session_per_task', True)

        self._prof.prof('task_run_start', uid=tid)
        task['proc'] = sp.Popen(args              = cmdline,
                                executable        = None,
                                shell             = False,
                                stdin             = None,
                                stdout            = _launch_out_h,
                                stderr            = sp.STDOUT,
                                start_new_session = _start_new_session,
                                close_fds         = True,
                                cwd               = sbox)
        self._prof.prof('task_run_ok', uid=tid)

        # store pid for last-effort termination
        _pids.append(task['proc'].pid)

        # handle task timeout if needed
        self.handle_timeout(task)

        # watch task for completion
        self._watch_queue.put([self.TO_WATCH, task])


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        to_watch  = list()  # contains task dicts
        to_cancel = set()   # contains task IDs

        try:
            while not self._term.is_set():

                # FIXME: we don't want to only wait for one Task -- then we
                #        would pull Task state too frequently.  OTOH, we
                #        also don't want to learn about tasks until all
                #        slots are filled, because then we may not be able
                #        to catch finishing tasks in time -- so there is
                #        a fine balance here.  Balance means 100.
                MAX_QUEUE_BULKSIZE = 100
                count = 0

                try:
                    while count < MAX_QUEUE_BULKSIZE:

                        flag, thing = self._watch_queue.get_nowait()
                        count += 1

                        # NOTE: `thing` can be task id or task dict, depending
                        #       on the flag value
                        if   flag == self.TO_WATCH : to_watch.append(thing)
                        elif flag == self.TO_CANCEL: to_cancel.add(thing)
                        else: raise RuntimeError('unknown flag %s' % flag)

                except queue.Empty:
                    # nothing found -- no problem, see if any tasks finished
                    pass

                # check on the known tasks.
                action = self._check_running(to_watch, to_cancel)

                # FIXME: remove uids from lists after completion

                if not action and not count:
                    # nothing happened at all!  Zzz for a bit.
                    # FIXME: make configurable
                    time.sleep(0.1)

        except Exception as e:
            self._log.exception('Error in ExecWorker watch loop (%s)' % e)
            # FIXME: this should signal the ExecWorker for shutdown...


    # --------------------------------------------------------------------------
    # Iterate over all running tasks, check their status, and decide on the
    # next step.  Also check for a requested cancellation for the tasks.
    def _check_running(self, to_watch, to_cancel):

        #
        action = False

        # `to_watch.remove()` in the loop requires copy to iterate over the list
        for task in list(to_watch):

            tid = task['uid']

            # poll subprocess object
            exit_code = task['proc'].poll()

            if exit_code is None:

                # process is still running - cancel if needed
                if tid in to_cancel:

                    self._log.debug('cancel %s', tid)

                    action = True
                    self._prof.prof('task_run_cancel_start', uid=tid)

                    # got a request to cancel this task - send SIGTERM to the
                    # process group (which should include the actual launch
                    # method)
                    try:
                        # kill the whole process group
                        pgrp = os.getpgid(task['proc'].pid)
                        os.killpg(pgrp, signal.SIGKILL)
                    except OSError:
                        # lost race: task is already gone, we ignore this
                        # FIXME: collect and move to DONE/FAILED
                        pass

                    task['proc'].wait()  # make sure proc is collected

                    to_cancel.remove(tid)
                    to_watch.remove(task)
                    del task['proc']  # proc is not json serializable

                    self._prof.prof('task_run_cancel_stop', uid=tid)

                    self._prof.prof('unschedule_start', uid=tid)
                    self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)
                    self.advance_tasks(task, rps.CANCELED, publish=True,
                                                           push=False)

            else:

                action = True
                self._prof.prof('task_run_stop', uid=tid)

                # make sure proc is collected
                task['proc'].wait()

                # we have a valid return code -- task is final
                self._log.info("Task %s has return code %s.", tid, exit_code)

                task['exit_code'] = exit_code

                # Free the Slots, Flee the Flots, Ree the Frots!
                to_watch.remove(task)
                if tid in to_cancel:
                    to_cancel.remove(tid)
                del task['proc']  # proc is not json serializable

                self._prof.prof('unschedule_start', uid=tid)
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

                if exit_code != 0:
                    # task failed - fail after staging output
                    task['exception']        = 'RuntimeError("task failed")'
                    task['exception_detail'] = 'exit code: %s' % exit_code
                    task['target_state'    ] = rps.FAILED

                else:
                    # The task finished cleanly, see if we need to deal with
                    # output data.  We always move to stageout, even if there
                    # are no directives -- at the very least, we'll upload
                    # stdout/stderr
                    task['target_state'] = rps.DONE

                self.advance_tasks(task, rps.AGENT_STAGING_OUTPUT_PENDING,
                                         publish=True, push=True)

        return action


    # --------------------------------------------------------------------------
    #
    # pylint: disable=unused-argument
    def _get_prof(self, event, tid, msg=''):

        return '$RP_PROF %s\n' % event


    # --------------------------------------------------------------------------
    #
    def _get_rp_funcs(self):

        # define helper functions
        ret  = '\nrp_error() {\n'
        ret += '    echo "$1 failed" 1>&2\n'
        ret += '    exit 1\n'
        ret += '}\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_rp_env(self, task):

        tid  = task['uid']
        name = task.get('name') or tid
        sbox = os.path.realpath(task['task_sandbox_path'])

        if sbox.startswith(self._pwd):
            sbox = '$RP_PILOT_SANDBOX%s' % sbox[len(self._pwd):]

        ret  = '\n'
        ret += 'export RP_TASK_ID="%s"\n'          % tid
        ret += 'export RP_TASK_NAME="%s"\n'        % name
        ret += 'export RP_PILOT_ID="%s"\n'         % self._pid
        ret += 'export RP_SESSION_ID="%s"\n'       % self.sid
        ret += 'export RP_RESOURCE="%s"\n'         % self.resource
        ret += 'export RP_RESOURCE_SANDBOX="%s"\n' % self.rsbox
        ret += 'export RP_SESSION_SANDBOX="%s"\n'  % self.ssbox
        ret += 'export RP_PILOT_SANDBOX="%s"\n'    % self.psbox
        ret += 'export RP_TASK_SANDBOX="%s"\n'     % sbox
        # FIXME AM
      # ret += 'export RP_LFS="%s"\n'              % self.lfs
        ret += 'export RP_GTOD="%s"\n'             % self.gtod
        ret += 'export RP_PROF="%s"\n'             % self.prof
      # ret += 'export RP_REGISTRY_URL="%s"\n'     % self.reg_addr

        if self._prof.enabled:
            ret += 'export RP_PROF_TGT="%s/%s.prof"\n' % (sbox, tid)
        else:
            ret += 'unset  RP_PROF_TGT\n'

        return ret


    # --------------------------------------------------------------------------
    #
    # launcher
    #
    def _get_launch_env(self, launcher):

        ret  = ''

        for cmd in launcher.get_launcher_env():
            ret += '%s || rp_error launcher_env\n' % cmd

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_prep_launch(self, task, sig):

        ret = ''
        td = task['description']

        if sig not in td:
            return ret

        for cmd in ru.as_list(task['description'][sig]):
            ret += '%s || rp_error %s\n' % (cmd, sig)

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_launch(self, task, launcher, exec_path):

        ret  = '( \\\n'

        for cmd in ru.as_list(launcher.get_launch_cmds(task, exec_path)):
            ret += '  %s \\\n' % cmd

        ret += ') 1> %s \\\n  2> %s\n' % (task['stdout_file_short'],
                                          task['stderr_file_short'])
        ret += 'RP_RET=$?\n'

        return ret


    # --------------------------------------------------------------------------
    #
    # exec
    #
    def _get_task_env(self, task, launcher):

        ret = ''
        td  = task['description']

        # named_env's are prepared by the launcher
        if td['named_env']:
            ret += '\n# named environment\n'
            ret += '. %s\n' % launcher.get_task_named_env(td['named_env'])

        # also add any env vars requested in the task description
        if td['environment']:
            ret += '\n# task env settings\n'
            for key, val in td['environment'].items():
                ret += 'export %s="%s"\n' % (key, val)

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_rank_ids(self, n_ranks, launcher):

        ret  = ''
        ret += 'export RP_RANKS=%s\n' % n_ranks
        ret += launcher.get_rank_cmd()

        if n_ranks > 1:

            # make sure that RP_RANK is known (otherwise task fails silently)
            if 'export RP_RANK=' not in ret:
                raise RuntimeError('launch method does not export RP_RANK')

        # also define a method to sync all ranks on certain events
        ret += '\nrp_sync_ranks() {\n'
        ret += '    sig=$1\n'
        ret += '    echo $RP_RANK >> $sig.sig\n'
        ret += '    while test $(cat $sig.sig | wc -l) -lt $RP_RANKS; do\n'
        ret += '        sleep 1\n'
        ret += '    done\n'
        ret += '}\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def _extend_pre_exec(self, td, ranks=None):

        # FIXME: this assumes that the rank has a `gpu_maps` and `core_maps`
        #        with exactly one entry, corresponding to the rank process to be
        #        started.

        # FIXME: need to distinguish between logical and physical IDs

        if td['threading_type'] == rpc.OpenMP:
            # for future updates: if task ranks are heterogeneous in terms of
            #                     number of threads, then the following string
            #                     should be converted into dictionary (per rank)
            num_threads = td.get('cores_per_rank', 1)
            td['pre_exec'].append('export OMP_NUM_THREADS=%d' % num_threads)

        if td['gpus_per_rank'] and td['gpu_type'] == rpc.CUDA and ranks:
            # equivalent to the 'physical' value for original `cvd_id_mode`
            rank_id  = 0
            rank_env = {}
            for slot_ranks in ranks:
                for gpu_map in slot_ranks['gpu_map']:
                    rank_env[str(rank_id)] = \
                        'export CUDA_VISIBLE_DEVICES=%s' % \
                        ','.join([str(g) for g in gpu_map])
                    rank_id += 1
            td['pre_exec'].append(rank_env)

        # pre-defined `pre_exec` per platform configuration
        td['pre_exec'].extend(ru.as_list(self._cfg.get('task_pre_exec')))


    # --------------------------------------------------------------------------
    #
    def _get_prep_exec(self, task, n_ranks, sig):

        ret = ''
        td  = task['description']

        if sig not in td:
            return ret

        entries         = ru.as_list(td[sig])
        switch_per_rank = any([isinstance(x, dict) for x in entries])
        cmd_template    = '%s || rp_error %s\n'

        sync_ranks_cmd = ''
        if sig == 'pre_exec' and td['pre_exec_sync']:
            sync_ranks_cmd = 'rp_sync_ranks %s\n' % sig

        if not switch_per_rank:
            return ''.join([cmd_template % (x, sig) for x in entries]) + \
                   sync_ranks_cmd

        ret += 'case "$RP_RANK" in\n'
        for rank_id in range(n_ranks):

            ret += '    %d)\n' % rank_id

            for entry in entries:

                if isinstance(entry, str):
                    entry = {str(rank_id): entry}

                for cmd in ru.as_list(entry.get(str(rank_id))):
                    ret += '        ' + cmd_template % (cmd, sig)

            ret += '        ;;\n'

        ret += 'esac\n' + sync_ranks_cmd

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_exec(self, task, launcher):

        # FIXME: core pinning goes here

        ret  = ''

        for cmd in ru.as_list(launcher.get_exec(task)):
            ret += '%s\n' % cmd
        ret += 'RP_RET=$?\n'

        return ret


# ------------------------------------------------------------------------------





import sys


# ------------------------------------------------------------------------------
#
# make sure deprecation warning is shown only once per type
#
_seen = list()


def _warn(old_type, new_type):
    if old_type not in _seen:
        _seen.append(old_type)
        sys.stderr.write('%s is deprecated - use %s\n' % (old_type, new_type))


# ------------------------------------------------------------------------------
#
from .task_description import TaskDescription


class ComputeUnitDescription(TaskDescription):

    def __init__(self, *args, **kwargs):

        _warn(ComputeUnitDescription, TaskDescription)
        TaskDescription.__init__(self, *args, **kwargs)


# ------------------------------------------------------------------------------
#
from .task import Task


class ComputeUnit(Task):

    def __init__(self, *args, **kwargs):

        _warn(ComputeUnit, Task)
        Task.__init__(self, *args, **kwargs)


# ------------------------------------------------------------------------------
#
from .pilot_description import PilotDescription


class ComputePilotDescription(PilotDescription):

    def __init__(self, *args, **kwargs):

        _warn(ComputePilotDescription, PilotDescription)
        PilotDescription.__init__(self, *args, **kwargs)


# ------------------------------------------------------------------------------
#
from .pilot import Pilot


class ComputePilot(Pilot):

    def __init__(self, *args, **kwargs):

        _warn(ComputePilot, Pilot)
        Pilot.__init__(self, *args, **kwargs)


# ------------------------------------------------------------------------------
#
from .task_manager import TaskManager


class UnitManager(TaskManager):

    def __init__(self, *args, **kwargs):

        _warn(UnitManager, TaskManager)
        TaskManager.__init__(self, *args, **kwargs)


# ------------------------------------------------------------------------------





import os
import glob
import tarfile

import radical.saga  as rs
import radical.utils as ru

from   .db_utils import get_session_docs

rs_fs = rs.filesystem


# ------------------------------------------------------------------------------
#
def fetch_json(sid, dburl=None, tgt=None, skip_existing=False, session=None,
               log=None):
    '''
    Returns:

        file name.

    '''

    if not log and session:
        log = session._log
    elif not log:
        log = ru.Logger('radical.pilot.utils')

    if session:
        rep = session._rep
    else:
        rep = ru.Reporter('radical.pilot.utils')

    if not tgt:
        tgt = os.getcwd()

    if tgt.startswith('/'):
        dst = '%s/%s/%s.json' % (tgt, sid, sid)
    else:
        dst = '%s/%s/%s/%s.json' % (os.getcwd(), tgt, sid, sid)

    ru.rec_makedir(os.path.dirname(dst))

    if skip_existing and os.path.isfile(dst) and os.path.getsize(dst):
        log.info("session already in %s", dst)
        return dst

    # need to fetch from MongoDB
    if not dburl:
        dburl = os.environ.get('RADICAL_PILOT_DBURL')

    if not dburl:
        raise ValueError('need RADICAL_PILOT_DBURL to fetch session')

    mongo, db, _, _, _ = ru.mongodb_connect(dburl)

    json_docs = get_session_docs(sid, db)
    ru.write_json(json_docs, dst)
    mongo.close()

    log.info("session written to %s", dst)
    rep.ok("+ %s (json)\n" % sid)

    return dst


# ------------------------------------------------------------------------------
#
def fetch_filetype(ext, name, sid, dburl=None, src=None, tgt=None, access=None,
        session=None, skip_existing=False, fetch_client=False, log=None):
    '''
    Args:

        - ext  (str): file extension to fetch
        - name (str): full name of filetype for log messages etc
        - sid  (str): session for which all files are fetched
        - src  (str): dir to look for client session files (`$src/$sid/*.ext`)
        - tgt  (str): dir to store the files in (`$tgt/$sid/*.ext`,
            `$tgt/$sid/$pid/*.ext`)

    Returns:

        list[str]: list of file names (fetched and/or cached)

    '''

    if not log and session:
        log = session._log

    elif not log:
        log = ru.Logger('radical.pilot.utils')

    if session:
        rep = session._rep
    else:
        rep = ru.Reporter('radical.pilot.utils')

    ret = list()

    if not src:
        src = os.getcwd()

    if not tgt:
        tgt = os.getcwd()

    # no point in fetching client files into client sandbox
    if src == tgt:
        fetch_client = False

    if not tgt.startswith('/') and '://' not in tgt:
        tgt = "%s/%s" % (os.getcwd(), tgt)

    # we always create a session dir as real target
    tgt_url = rs.Url("%s/%s/" % (tgt, sid))

    # turn URLs without `schema://host` into `file://localhost`,
    # so that they dont become interpreted as relative paths.
    if not tgt_url.schema: tgt_url.schema = 'file'
    if not tgt_url.host  : tgt_url.host   = 'localhost'

    # create target dir for session
    ru.rec_makedir(tgt_url.path)

    # first fetch client files
    if fetch_client:
        client_files = glob.glob("%s/%s/**.%s" % (src, sid, ext))
        if not client_files:
            raise RuntimeError('no client %s in %s/%s' % (name, src, sid))

        for client_file in client_files:

            ftgt = rs.Url('%s/%s' % (tgt_url, os.path.basename(client_file)))
            ret.append("%s" % ftgt.path)

            if skip_existing and os.path.isfile(ftgt.path) \
                             and os.path.getsize(ftgt.path):
                pass
            else:
                log.debug('fetch client file %s', client_file)
                rs_file = rs_fs.File(client_file, session=session)
                rs_file.copy(ftgt, flags=rs_fs.CREATE_PARENTS)
                rs_file.close()

    # we need the session json for pilot details
    json_name  = fetch_json(sid, dburl, tgt, skip_existing, session, log)
    json_docs  = ru.read_json(json_name)
    pilots     = json_docs['pilot']
    num_pilots = len(pilots)

    log.debug("Session: %s", sid)
    log.debug("Number of pilots in session: %d", num_pilots)

    for pilot in pilots:

        pid = pilot['uid']

        # create target dir for this pilot
        ru.rec_makedir('%s/%s' % (tgt_url.path, pid))

        try:
            log.debug("processing pilot '%s'", pid)

            sandbox_url = rs.Url(pilot['pilot_sandbox'])

            if access:
                # Allow to use a different access schema than used for the the
                # run.  Useful if you ran from the headnode, but would like to
                # retrieve the files to your desktop (Hello Titan).
                access_url = rs.Url(access)
                sandbox_url.schema = access_url.schema
                sandbox_url.host   = access_url.host

            sandbox = rs_fs.Directory (sandbox_url, session=session)

            # Try to fetch a tarball of files, so that we can get them
            # all in one (SAGA) go!
            tarball_name  = '%s.%s.tbz'   % (pid, ext)
            tarball_tgt   = '%s/%s/%s/%s' % (tgt, sid, pid, tarball_name)
            tarball_local = False

            # check if we have a local tarball already
            if skip_existing and \
               os.path.isfile(tarball_tgt) and \
               os.path.getsize(tarball_tgt):
                tarball_local = True

            if not tarball_local:
                # need to fetch tarball
                #  - if no remote tarball exists, create it
                #  - fetch remote tarball

                tarball_remote = False
                if sandbox.is_file(tarball_name) and \
                        sandbox.get_size(tarball_name):
                    tarball_remote = True

                if not tarball_remote:
                    # so lets create a tarball with SAGA JobService
                    js_url = pilot['js_hop']
                    log.debug('js  : %s', js_url)
                    js  = rs.job.Service(js_url, session=session)
                    cmd = "cd %s; find . -name \\*.%s > %s.lst; " \
                          "tar cjf %s -T %s.lst" % (sandbox.url.path, ext, ext,
                              tarball_name, ext)
                    j = js.run_job(cmd)
                    j.wait()

                    log.debug('tar cmd   : %s', cmd)
                    log.debug('tar result: %s\n---\n%s\n---\n%s',
                              j.get_stdout_string(), j.get_stderr_string(),
                              j.exit_code)

                    if j.exit_code:
                        raise RuntimeError('could not create tarball: %s' %
                                j.get_stderr_string())

                # we not have a remote tarball and can fetch it
                log.info("fetch '%s%s' to '%s'.", sandbox_url,
                         tarball_name, tgt_url)

                rs_file = rs_fs.File("%s%s" % (sandbox_url, tarball_name),
                                     session=session)
                rs_file.copy(tarball_tgt, flags=rs_fs.CREATE_PARENTS)
                rs_file.close()

            # we now have a local tarball - unpack it
            # note that we do not check if it was unpacked before - it's simpler
            # (and possibly cheaper) to just do that again
            log.info('Extract tarball %s', tarball_tgt)
            tarball = tarfile.open(tarball_tgt, mode='r:bz2')
            tarball.extractall("%s/%s" % (tgt_url.path, pid))

            files = glob.glob("%s/%s/**.%s" % (tgt_url.path, pid, ext))
            ret.extend(files)

            rep.ok("+ %s (%s)\n" % (pid, name))

        except Exception:
            # do not raise, we still try the other pilots
            rep.error("- %s (%s)\n" % (pid, name))
            log.exception('failed to fetch %s for %s', pid, name)

    return ret


# ------------------------------------------------------------------------------
#
def fetch_profiles (sid, dburl=None, src=None, tgt=None, access=None,
        session=None, skip_existing=False, fetch_client=False, log=None):

    return fetch_filetype('prof', 'profiles', sid, dburl, src, tgt, access,
            session, skip_existing, fetch_client, log)


# ------------------------------------------------------------------------------
#
def fetch_logfiles (sid, dburl=None, src=None, tgt=None, access=None,
        session=None, skip_existing=False, fetch_client=False, log=None):

    return fetch_filetype('log', 'logfiles', sid, dburl, src, tgt, access,
            session, skip_existing, fetch_client, log)


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2016-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import math
import os

from typing import Optional, List, Tuple, Dict, Any

T_NODES     = List[Tuple[str, int]]
T_NODE_LIST = List[Dict[str, Any]]


import radical.utils as ru

from ... import agent     as rpa
from ... import constants as rpc


# 'enum' for resource manager types
RM_NAME_FORK        = 'FORK'
RM_NAME_CCM         = 'CCM'
RM_NAME_LSF         = 'LSF'
RM_NAME_PBSPRO      = 'PBSPRO'
RM_NAME_SLURM       = 'SLURM'
RM_NAME_TORQUE      = 'TORQUE'
RM_NAME_COBALT      = 'COBALT'
RM_NAME_YARN        = 'YARN'
RM_NAME_DEBUG       = 'DEBUG'


# ------------------------------------------------------------------------------
#
class RMInfo(ru.TypedDict):
    '''
    Each resource manager instance must gather provide the information defined
    in this class.  Additional attributes can be attached, but should then only
    be evaluated by launch methods which are tightly bound to the resource
    manager type ('friends' in C++ speak).
    '''

    _schema = {
            'requested_nodes'  : int,           # number of requested nodes
            'requested_cores'  : int,           # number of requested cores
            'requested_gpus'   : int,           # number of requested gpus

            'partitions'       : {int: None},   # partition setup
            'node_list'        : [None],        # tuples of node uids and names
            'agent_node_list'  : [None],        # nodes reserved for sub-agents
            'service_node_list': [None],        # nodes reserved for services

            'cores_per_node'   : int,           # number of cores per node
            'threads_per_core' : int,           # number of threads per core

            'gpus_per_node'    : int,           # number of gpus per node
            'threads_per_gpu'  : int,           # number of threads per gpu
            'mem_per_gpu'      : int,           # memory per gpu (MB)

            'lfs_per_node'     : int,           # node local FS size (MB)
            'lfs_path'         : str,           # node local FS path
            'mem_per_node'     : int,           # memory per node (MB)

            'details'          : {None: None},  # dict of launch method info
            'lm_info'          : {str: None},   # dict of launch method info
    }

    _defaults = {
            'agent_node_list'  : [],            # no sub-agents run by default
            'service_node_list': [],            # no services run by default
            'cores_per_node'   : 1,
            'threads_per_core' : 1,
            'gpus_per_node'    : 0,
            'threads_per_gpu'  : 1,
    }


    # --------------------------------------------------------------------------
    #
    def _verify(self):

        assert self['requested_nodes'  ]
        assert self['requested_cores'  ]
        assert self['requested_gpus'   ] is not None

      # assert self['partitions'       ] is not None
        assert self['node_list']
        assert self['agent_node_list'  ] is not None
        assert self['service_node_list'] is not None

        assert self['cores_per_node'   ]
        assert self['gpus_per_node'    ] is not None
        assert self['threads_per_core' ] is not None


# ------------------------------------------------------------------------------
#
# Base class for ResourceManager implementations.
#
class ResourceManager(object):
    '''
    The Resource Manager provides fundamental resource information via
    `self.info` (see `RMInfo` class definition).

      ResourceManager.node_list      : list of nodes names and uids
      ResourceManager.agent_node_list: list of nodes reserved for agent procs
      ResourceManager.cores_per_node : number of cores each node has available
      ResourceManager.gpus_per_node  : number of gpus  each node has available

    Schedulers can rely on these information to be available.  Specific
    ResourceManager incarnation may have additional information available -- but
    schedulers relying on those are invariably bound to the specific
    ResourceManager.

    The ResourceManager will reserve nodes for the agent execution, by deriving
    the respectively required node count from the config's 'agents' section.
    Those nodes will be listed in ResourceManager.agent_node_list. Schedulers
    MUST NOT use the agent_node_list to place tasks -- Tasks are limited
    to the nodes in ResourceManager.node_list.

    Last but not least, the RM will initialize launch methods and ensure that
    the executor (or any other component really) finds them ready to use.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, log, prof):

        self.name  = type(self).__name__
        self._cfg  = cfg
        self._log  = log
        self._prof = prof

        self._log.info('Configuring ResourceManager %s', self.name)

        reg     = ru.zmq.RegistryClient(url=self._cfg.reg_addr)
        rm_info = reg.get('rm.%s' % self.name.lower())

        if rm_info:

            self._log.debug('RM init from registry')
            rm_info = RMInfo(rm_info)
            rm_info.verify()

        else:
            self._log.debug('RM init from scratch')

            # let the base class collect some data, then let the impl take over
            rm_info = self.init_from_scratch()
            rm_info.verify()

            # have a valid info - store in registry and complete initialization
            reg.put('rm.%s' % self.name.lower(), rm_info.as_dict())

        # set up launch methods even when initialized from registry info
        self._prepare_launch_methods(rm_info)

        reg.close()
        self._set_info(rm_info)


    # --------------------------------------------------------------------------
    #
    @property
    def info(self):

        return self._rm_info


    # --------------------------------------------------------------------------
    #
    def _set_info(self, info):

        self._rm_info = info
        self._rm_info.verify()


    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:
        '''
        This method MUST be overloaded by any RM implementation.  It will be
        called during `init_from_scratch` and is expected to check and correct
        or complete node information, such as `cores_per_node`, `gpus_per_node`
        etc., and to provide `rm_info.node_list` of the following form:

            node_list = [
                {
                    'node_name': str                        # node name
                    'node_id'  : str                        # node uid
                    'cores'    : [rpc.FREE, rpc.FREE, ...]  # cores per node
                    'gpus'     : [rpc.FREE, rpc.FREE, ...]  # gpus per node
                    'lfs'      : int                        # lfs per node (MB)
                    'mem'      : int                        # mem per node (MB)
                },
                ...
            ]

        The node entries can be augmented with additional information which may
        be interpreted by the specific agent scheduler instance.
        '''

        raise NotImplementedError('_init_from_scratch is not implemented')


    # --------------------------------------------------------------------------
    #
    def init_from_scratch(self):

        rm_info = RMInfo()

        # fill well defined default attributes
        rm_info.requested_nodes  = self._cfg.nodes
        rm_info.requested_cores  = self._cfg.cores
        rm_info.requested_gpus   = self._cfg.gpus
        rm_info.cores_per_node   = self._cfg.cores_per_node
        rm_info.gpus_per_node    = self._cfg.gpus_per_node
        rm_info.lfs_per_node     = self._cfg.lfs_size_per_node
        rm_info.lfs_path         = ru.expand_env(self._cfg.lfs_path_per_node)

        rm_info.threads_per_gpu  = 1
        rm_info.mem_per_gpu      = None

        rcfg                     = self._cfg.resource_cfg
        rm_info.mem_per_node     = rcfg.mem_per_node or 0

        system_architecture      = rcfg.get('system_architecture', {})
        rm_info.threads_per_core = int(os.environ.get('RADICAL_SMT') or
                                       system_architecture.get('smt', 1))

        # let the specific RM instance fill out the RMInfo attributes
        rm_info = self._init_from_scratch(rm_info)
        alloc_nodes = len(rm_info.node_list)

        # we expect to have a valid node list now
        self._log.info('node list: %s', rm_info.node_list)

        # the config can override core and gpu detection,
        # and decide to block some resources (part of the core specialization)
        blocked_cores = system_architecture.get('blocked_cores', [])
        blocked_gpus  = system_architecture.get('blocked_gpus',  [])

        self._log.info('blocked cores: %s' % blocked_cores)
        self._log.info('blocked gpus : %s' % blocked_gpus)

        if blocked_cores or blocked_gpus:

            rm_info.cores_per_node -= len(blocked_cores)
            rm_info.gpus_per_node  -= len(blocked_gpus)

            for node in rm_info.node_list:

                for idx in blocked_cores:
                    assert len(node['cores']) > idx
                    node['cores'][idx] = rpc.DOWN

                for idx in blocked_gpus:
                    assert len(node['gpus']) > idx
                    node['gpus'][idx] = rpc.DOWN

        # number of nodes could be unknown if `cores_per_node` is not in config,
        # but is provided by a corresponding RM in `_init_from_scratch`
        if not rm_info.requested_nodes:
            n_nodes = rm_info.requested_cores / rm_info.cores_per_node
            if rm_info.gpus_per_node:
                n_nodes = max(
                    rm_info.requested_gpus / rm_info.gpus_per_node,
                    n_nodes)
            rm_info.requested_nodes = math.ceil(n_nodes)

        assert alloc_nodes                          >= rm_info.requested_nodes
        assert alloc_nodes * rm_info.cores_per_node >= rm_info.requested_cores
        assert alloc_nodes * rm_info.gpus_per_node  >= rm_info.requested_gpus

        # The ResourceManager may need to reserve nodes for sub agents and
        # service, according to the agent layout and pilot config.  We dig out
        # the respective requirements from the node list and complain on
        # insufficient resources
        agent_nodes   = 0
        service_nodes = 0

        for agent in self._cfg.get('agents', {}):
            if self._cfg['agents'][agent].get('target') == 'node':
                agent_nodes += 1

        if os.path.isfile('./services'):
            service_nodes += 1

        # Check if the ResourceManager implementation reserved agent nodes.
        # If not, pick the first couple of nodes from the nodelist as fallback.
        if agent_nodes:

            if not rm_info.agent_node_list:
                for _ in range(agent_nodes):
                    rm_info.agent_node_list.append(rm_info.node_list.pop())

            assert agent_nodes == len(rm_info.agent_node_list)

        if service_nodes:

            if not rm_info.service_node_list:
                for _ in range(service_nodes):
                    rm_info.service_node_list.append(rm_info.node_list.pop())

            assert service_nodes == len(rm_info.service_node_list)

        self._log.info('compute nodes: %s' % len(rm_info.node_list))
        self._log.info('agent   nodes: %s' % len(rm_info.agent_node_list))
        self._log.info('service nodes: %s' % len(rm_info.service_node_list))

        # check if we can do any work
        if not rm_info.node_list:
            raise RuntimeError('ResourceManager has no nodes left to run tasks')

        return rm_info


    # --------------------------------------------------------------------------
    #
    def _prepare_launch_methods(self, rm_info):

        launch_methods  = self._cfg.resource_cfg.launch_methods

        self._launchers    = {}
        self._launch_order = launch_methods.get('order') or list(launch_methods)

        for lm_name in list(self._launch_order):

            lm_cfg = launch_methods[lm_name]

            try:
                self._log.debug('prepare lm %s', lm_name)
                lm_cfg['pid']         = self._cfg.pid
                lm_cfg['reg_addr']    = self._cfg.reg_addr
                lm_cfg['resource']    = self._cfg.resource
                self._launchers[lm_name] = rpa.LaunchMethod.create(
                    lm_name, lm_cfg, rm_info, self._log, self._prof)

            except:
                self._log.exception('skip lm %s', lm_name)
                self._launch_order.remove(lm_name)

        if not self._launchers:
            raise RuntimeError('no valid launch methods found')


    # --------------------------------------------------------------------------
    #
    def get_partitions(self):

        # TODO: this makes it impossible to have mutiple launchers with a notion
        #       of partitions

        for lname in self._launchers:

            launcher   = self._launchers[lname]
            partitions = launcher.get_partitions()

            if partitions:
                return partitions


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the
    # ResourceManager.
    #
    @classmethod
    def create(cls, name, cfg, log, prof):

        from .ccm         import CCM
        from .fork        import Fork
        from .lsf         import LSF
        from .pbspro      import PBSPro
        from .slurm       import Slurm
        from .torque      import Torque
        from .cobalt      import Cobalt
        from .yarn        import Yarn
        from .debug       import Debug

        # Make sure that we are the base-class!
        if cls != ResourceManager:
            raise TypeError('ResourceManager Factory only available to base class!')

        impl = {
            RM_NAME_FORK        : Fork,
            RM_NAME_CCM         : CCM,
            RM_NAME_LSF         : LSF,
            RM_NAME_PBSPRO      : PBSPro,
            RM_NAME_SLURM       : Slurm,
            RM_NAME_TORQUE      : Torque,
            RM_NAME_COBALT      : Cobalt,
            RM_NAME_YARN        : Yarn,
            RM_NAME_DEBUG       : Debug
        }

        if name not in impl:
            raise RuntimeError('ResourceManager %s unknown' % name)

        return impl[name](cfg, log, prof)



    # --------------------------------------------------------------------------
    #
    def stop(self):

        # clean up launch methods
        for name in self._launchers:
            try:    self._launchers[name].finalize()
            except: self._log.exception('LM %s finalize failed', name)


    # --------------------------------------------------------------------------
    #
    def find_launcher(self, task):

        errors = list()
        for name in self._launch_order:

            launcher = self._launchers[name]
            lm_can_launch, err_message = launcher.can_launch(task)
            if lm_can_launch:
                return launcher
            else:
                errors.append([name, err_message])

        self._log.error('no launch method for task %s:', task['uid'])
        for name, error in errors:
            self._log.debug('    %s: %s', name, error)

        return None


    # --------------------------------------------------------------------------
    #
    def _parse_nodefile(self, fname: str,
                              cpn  : Optional[int] = 0,
                              smt  : Optional[int] = 1) -> T_NODES:
        '''
        parse the given nodefile and return a list of tuples of the form

            [['node_1', 42 * 4],
             ['node_2', 42 * 4],
             ...
            ]

        where the first tuple entry is the name of the node found, and the
        second entry is the number of entries found for this node.  The latter
        number usually corresponds to the number of process slots available on
        that node.

        Some nodefile formats though have one entry per node, not per slot.  In
        those cases we'll use the passed cores per node (`cpn`) to fill the slot
        count for the returned node list (`cpn` will supercede the detected slot
        count).

        An invalid or un-parsable file will result in an empty list being
        returned.
        '''

        if not smt:
            smt = 1

        self._log.info('using nodefile: %s', fname)
        try:
            nodes = dict()
            with ru.ru_open(fname, 'r') as fin:
                for line in fin.readlines():
                    node = line.strip()
                    assert ' ' not in node
                    if node in nodes: nodes[node] += 1
                    else            : nodes[node]  = 1

            if cpn:
                for node in list(nodes.keys()):
                    nodes[node] = cpn

            # convert node dict into tuple list
            return [(node, cpn * smt) for node, cpn in nodes.items()]

        except Exception:
            return []


    # --------------------------------------------------------------------------
    #
    def _get_cores_per_node(self, nodes: T_NODES) -> Optional[int]:
        '''
        From a node dict as returned by `self._parse_nodefile()`, determine the
        number of cores per node.  To do so, we check if all nodes have the same
        number of cores.  If that is the case we return that number.  If the
        node list is heterogeneous we will raise an `ValueError`.
        '''

        cores_per_node = set([node[1] for node in nodes])

        if len(cores_per_node) == 1:
            cores_per_node = cores_per_node.pop()
            self._log.debug('found %d [%d cores]', len(nodes), cores_per_node)
            return cores_per_node

        else:
            raise ValueError('non-uniform node list, cores_per_node invalid')


    # --------------------------------------------------------------------------
    #
    def _get_node_list(self, nodes  : T_NODES,
                             rm_info: RMInfo) -> T_NODE_LIST:
        '''
        From a node dict as returned by `self._parse_nodefile()`, and from
        additonal per-node information stored in rm_info, create a node list
        as required for rm_info.
        '''

        # keep nodes to be indexed (node_id) starting at 1
        # (required for jsrun ERF spec files and expanded to all other RMs)
        node_list = [{'node_name': node[0],
                      'node_id'  : str(idx + 1),
                      'cores'    : [rpc.FREE] * node[1],
                      'gpus'     : [rpc.FREE] * rm_info.gpus_per_node,
                      'lfs'      : rm_info.lfs_per_node,
                      'mem'      : rm_info.mem_per_node}
                     for idx, node in enumerate(nodes)]

        return node_list


# ------------------------------------------------------------------------------





# pylint: disable=global-statement

import os
import time

from typing import Union

import radical.utils as ru

# max number of t out/err chars to push to tail
MAX_IO_LOGLENGTH = 1024

# cache resource configs
_rcfgs = None


# ------------------------------------------------------------------------------
#
def tail(txt: str, maxlen: int = MAX_IO_LOGLENGTH) -> str:

    # shorten the given string to the last <n> characters, and prepend
    # a notification.  This is used to keep logging information in mongodb
    # manageable(the size of mongodb documents is limited).

    if not txt:
        return ''

    if len(txt) > maxlen:
        return "[... CONTENT SHORTENED ...]\n%s" % txt[-maxlen:]
    else:
        return txt


# ------------------------------------------------------------------------------
#
def get_rusage() -> str:

    import resource

    self_usage  = resource.getrusage(resource.RUSAGE_SELF)
    child_usage = resource.getrusage(resource.RUSAGE_CHILDREN)

    rtime = time.time()
    utime = self_usage.ru_utime  + child_usage.ru_utime
    stime = self_usage.ru_stime  + child_usage.ru_stime
    rss   = self_usage.ru_maxrss + child_usage.ru_maxrss

    return "real %3f sec | user %.3f sec | system %.3f sec | mem %.2f kB" \
         % (rtime, utime, stime, rss)


# ------------------------------------------------------------------------------
#
def create_tar(tgt: str, dnames: str) -> None:
    '''
    Create a tarball on the file system which contains all given directories
    '''

    uid   = os.getuid()
    gid   = os.getgid()
    mode  = 16893
    mtime = int(time.time())

    fout  = open(tgt, 'wb')

    def rpad(s, size):
        return s + (size - len(s)) * '\0'

    def write_dir(path):
        data  = rpad(path, 100) \
              + rpad('%o' % mode,   8) \
              + rpad('%o' % uid,    8) \
              + rpad('%o' % gid,    8) \
              + rpad('%o' % 0,     12) \
              + rpad('%o' % mtime, 12) \
              + 8 * '\0' + '5'
        cksum = 256 + sum(ord(h) for h in data)
        data  = rpad(data  , 512)
        data  = data  [:-364] + '%06o\0' % cksum + data[-357:]
        fout.write(ru.as_bytes(data))

    for dname in dnames:
        write_dir(dname)

    fout.close()


# ------------------------------------------------------------------------------
#
def get_resource_configs() -> ru.Config:
    '''
    Return all resource configurations used by `radical.pilot`.

    Configurations for the individual resources are organized as sites and
    resources:

       - cfgs = get_resource_configs()
       - sites = cfgs.keys()
       - for site in sites: resource_names = cfgs[site].keys()

    Returns:

        :obj:`radical.utils.Config`: the resource configurations
    '''

    global _rcfgs

    if not _rcfgs:
        _rcfgs = ru.Config('radical.pilot.resource', name='*', expand=False)

    # return a deep copy
    return ru.Config(from_dict=_rcfgs)


# ------------------------------------------------------------------------------
#
def get_resource_config(resource: str) -> Union[None, ru.Config]:
    '''
    For the given resource label, return the resource configuration used by
    `radical.pilot`.


    Args:
        resource (:obj:`str`): resource label for which to return the cfg

    Returns:

        :obj:`radical.utils.Config`: the resource configuration

        The method returns `None` if no resource config is found for the
        specified resource label.
    '''

    # populate cache
    if not _rcfgs:
        get_resource_configs()

    site, host = resource.split('.', 1)

    if site not in _rcfgs:
        return None

    if host not in _rcfgs[site]:
        return None

    # return a deep copy
    return ru.Config(cfg=_rcfgs[site][host])


# ------------------------------------------------------------------------------
#
def get_resource_fs_url(resource: str,
                        schema  : str = None) -> Union[None, ru.Url]:
    '''
    For the given resource label, return the contact URL of the resource's file
    system.  This corresponds to the `filesystem_endpoint` setting in the
    resource config.

    For example,
    `rs.filesystem.directory(get_resource_fs_url(...)).change_dir('/')`
    is equivalent to the base ``endpoint:///`` URL available for use
    in a `staging_directive`.

    Args:
        resource (:obj:`str`): resource label for which to return the url

        schema (:obj:`str`, optional): access schema to use for resource
            access.  Defaults to the default access schema as defined in the
            resources config files.

    Returns:

        :obj:`radical.utils.Url`: the file system URL

        The method returns `None` if no resource config is found for the
        specified resource label and access schema.
    '''

    rcfg = get_resource_config(resource)

    if not schema:
        schema = rcfg['schemas'][0]

    # return a deep copy
    return ru.Url(rcfg[schema]['filesystem_endpoint'])


# ------------------------------------------------------------------------------
#
def get_resource_job_url(resource: str,
                         schema  : str = None) -> Union[None, ru.Url]:
    '''
    For the given resource label, return the contact URL of the resource's job
    manager.


    Args:
        resource (:obj:`str`): resource label for which to return the url

        schema (:obj:`str`, optional): access schema to use for resource
            access.  Defaults to the default access schema as defined in the
            resources config files.

    Returns:

        :obj:`radical.utils.Url`: the job manager URL

        The method returns `None` if no resource config is found for the
        specified resource label and access schema.
    '''

    rcfg = get_resource_config(resource)

    if not schema:
        schema = rcfg['schemas'][0]

    # return a deep copy
    return ru.Url(rcfg[schema]['job_manager_endpoint'])


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import radical.pilot as rp
import radical.utils as ru

verbose  = os.environ.get('RADICAL_PILOT_VERBOSE', 'REPORT')
os.environ['RADICAL_PILOT_VERBOSE'] = verbose

""" DESCRIPTION: Tutorial 1: A Simple Workload consisting of a Bag-of-Tasks
"""

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


#
if __name__ == "__main__":

    RESOURCE_LABEL = 'local.localhost'
    PILOT_CORES    =  2
    BAG_SIZE       = 10  # The number of tasks
    TASK_CORES     =  1  # The cores each Task will take.
    QUEUE          = None
    # we use a reporter class for nicer output
    report = ru.Reporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' in the 'finally' clause.
    try:

        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        pmgr = rp.PilotManager(session=session)

        #
        # Change the resource below if you want to run on a remote resource.
        # You also might have to set the 'project' to your allocation ID if
        # your remote resource does compute time accounting.
        #
        # A list of preconfigured resources can be found at:
        # https://radicalpilot.readthedocs.io/en/stable/ \
        #        machconf.html#preconfigured-resources
        #
        pdesc = rp.PilotDescription ()
        pdesc.resource = RESOURCE_LABEL  # this is a "label", not a hostname
        pdesc.cores    =  PILOT_CORES
        pdesc.runtime  = 30    # minutes
        pdesc.cleanup  = True  # clean pilot sandbox and database entries
        pdesc.queue = QUEUE

        # submit the pilot.
        report.header("Submitting  Pilot to Pilot Manager ...")
        pilot = pmgr.submit_pilots(pdesc)

        # create a TaskManager which schedules Tasks over pilots.
        report.header("Initializing Task Manager ...")
        tmgr = rp.TaskManager (session=session)


        # Add the created Pilot to the TaskManager.
        report.ok('>>ok\n')

        tmgr.add_pilots(pilot)

        report.info('Create %d Task Description(s)\n\t' % BAG_SIZE)

        # create Task descriptions
        taskdesc_list = []
        for i in range(BAG_SIZE):

            # -------- BEGIN USER DEFINED Task DESCRIPTION --------- #
            taskdesc = rp.TaskDescription()
            taskdesc.executable  = "/bin/echo"
            taskdesc.arguments   = ['I am Task number $TASK_NO']
            taskdesc.environment = {'TASK_NO': i}
            taskdesc.cores       = 1
            # -------- END USER DEFINED Task DESCRIPTION --------- #

            taskdesc_list.append(taskdesc)
            report.progress()
        report.ok('>>>ok\n')
        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        report.header("Submit Tasks to Task Manager ...")
        task_set = tmgr.submit_tasks (taskdesc_list)

        report.header("Waiting for tasks to complete ...")
        tmgr.wait_tasks()

        for task in task_set:
            print("* Task %s, state %s, exit code: %s, stdout: %s"
                % (task.uid, task.state, task.exit_code, task.stdout.strip()))


    except Exception as e:
        # Something unexpected happened in the pilot code above
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        report.header("closing session")
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots.
    report.header()


# ------------------------------------------------------------------------------





import os
import sys

from typing import Dict, List, Any, Union

import radical.utils as ru

from .constants import DEFAULT_ACTION, DEFAULT_FLAGS, DEFAULT_PRIORITY


# ------------------------------------------------------------------------------
#
def expand_description(descr: Dict[str, Any]) -> None:
    """Get description dictionary for simple string directive.

    Convert any simple, string based staging directive in the given task
    description into its dictionary equivalent.

    In this context, the following kinds of expansions are performed::
        in:  ['input.dat']
        out: {'source' : 'client:///input.dat',
            'target' : 'task:///input.dat',
            'action' : rp.TRANSFER}

        in:  ['input.dat > staged.dat']
        out: {'source' : 'client:///input.dat',
            'target' : 'task:///staged.dat',
            'action' : rp.TRANSFER}

    This method changes the given description in place - repeated calls on the
    same description instance will have no effect.  However, we expect this
    method to be called only once during task construction.
    """

    if descr.get('input_staging')  is None: descr['input_staging']  = list()
    if descr.get('output_staging') is None: descr['output_staging'] = list()

    descr['input_staging' ] = expand_staging_directives(descr['input_staging' ])
    descr['output_staging'] = expand_staging_directives(descr['output_staging'])


# ------------------------------------------------------------------------------
#
def expand_staging_directives(sds: Union[str, Dict[str, Any], List[str]]
                             ) -> List[Dict[str, Any]]:
    """Take an abbreviated or compressed staging directive and expand it."""

    if not sds:
        return []

    sds = ru.as_list(sds)
    ret = list()
    for sd in sds:

        if isinstance(sd, str):
            # We detected a string, convert into dict.  The interpretation
            # differs depending of redirection characters being present in the
            # string.

            if   '>>' in sd: src, tgt = sd.split('>>', 2)
            elif '>'  in sd: src, tgt = sd.split('>' , 2)
            elif '<<' in sd: tgt, src = sd.split('<<', 2)
            elif '<'  in sd: tgt, src = sd.split('<' , 2)
            else           : src, tgt = sd, os.path.basename(ru.Url(sd).path)

            # FIXME: ns = session ID
            expanded = {
                    'uid'             : ru.generate_id('sd', ru.ID_SIMPLE),
                    'source'          : src.strip(),
                    'target'          : tgt.strip(),
                    'action'          : DEFAULT_ACTION,
                    'flags'           : DEFAULT_FLAGS,
                    'priority'        : DEFAULT_PRIORITY,
                    'exception'       : None,
                    'exception_detail': None
            }

        elif isinstance(sd, dict):

            # sanity check on dict syntax
            valid_keys = ['source', 'target', 'action', 'flags', 'priority',
                          'uid', 'prof_id']
            for k in sd.keys():
                if k not in valid_keys:
                    raise ValueError('"%s" is invalid on staging directive' % k)

            source   = sd.get('source')
            target   = sd.get('target',   os.path.basename(ru.Url(source).path))
            action   = sd.get('action',   DEFAULT_ACTION)
            flags    = sd.get('flags',    DEFAULT_FLAGS)
            priority = sd.get('priority', DEFAULT_PRIORITY)

            if not source:
                raise Exception("Staging directive dict has no source member!")

            # RCT flags should always be rendered as OR'ed integers - but old
            # versions of the RP API rendered them as list of strings.  We
            # convert to the integer version for backward compatibility - but we
            # complain loudly if we find actual strings.
            if isinstance(flags, list):
                int_flags = 0
                for flag in flags:
                    if isinstance(flags, str):
                        raise ValueError('"%s" is no valid RP constant' % flag)
                    int_flags |= flag
                flags = int_flags

            elif isinstance(flags, str):
                raise ValueError('use RP constants for staging flags!')

            expanded = {
                    'uid'             : ru.generate_id('sd'),
                    'source'          : source,
                    'target'          : target,
                    'action'          : action,
                    'flags'           : flags,
                    'priority'        : priority,
                    'exception'       : None,
                    'exception_detail': None
            }

        else:
            raise Exception("Unknown directive: %s (%s)" % (sd, type(sd)))

        # we warn the user when src or tgt are using the deprecated
        # `staging://` schema
        if str(expanded['source']).startswith('staging://'):
            sys.stderr.write('staging:// schema is deprecated - use pilot://\n')
            expanded['source'] = str(expanded['source']).replace('staging://',
                                                                 'pilot://')

        if str(expanded['target']).startswith('staging://'):
            sys.stderr.write('staging:// schema is deprecated - use pilot://\n')
            expanded['target'] = str(expanded['target']).replace('staging://',
                                                                 'pilot://')
        ret.append(expanded)

    return ret


# ------------------------------------------------------------------------------
#
def complete_url(path   : str,
                 context: Dict[str, str],
                 log    : ru.Logger = None
                ) -> ru.Url:
    """Construct a URL.

    Some paths in data staging directives are to be interpreted relative to
    certain locations, namely relative to

    * `client://`  : the client's working directory
    * `endpoint://`: the root of the target resource's file system
    * `resource://`: the sandbox base dir on the target resource
    * `session://` : the session sandbox  on the target resource
    * `pilot://`   : the pilot   sandbox  on the target resource
    * `task://`    : the task    sandbox  on the target resource

    All location are interpreted as directories, never as files.

    For the above schemas, we interpret `schema://` the same as `schema:///`,
    ie. we treat this as a namespace, not as location qualified by a hostname.

    The `context` parameter is expected to be a dict which provides a set of
    URLs to be used to expand the path for those schemas

    Other URL schemas are left alone, any non-URL strings are interpreted as
    path in the context of `pwd`, i.e. in the context of the working directory
    of the current process, unless a `pwd` entry is provided in the `context`
    dict). `file://` schemas are left unaltered and are expected to point to
    absolute locations in the file system.

    The method returns an instance of `:py:class:`radical.utils.Url`.  Even if
    the URL is not altered, a new instance (deep copy) will be returned.

    `endpoint://` is based on the `filesystem_endpoint` attribute of the
    resource config and points to the file system accessible  via that URL.
    Note that the notion of 'root' dependends of the access protocol and the
    providing service implementation.

    Note:
        URL parsing is not really cheap, so this method should be used
        conservatively.

    """

    # FIXME: consider evaluation of env vars

    purl     = ru.Url(path)
    str_path = str(path)

    # we always want a schema, and fall back to file:// or pwd://, depending if
    # the path is absolute or relative.  Note that pwd:// is interpreted in the
    # context of the staging component, which may live on the client or target
    # resource -- we make no attempt at expanding the path at this point.
    # We further assume that the user knows what she is doing when using
    # absolute paths, and make no attempts to verify those either.
    if not purl.schema:
        if str_path.startswith('/'):
            purl.schema = 'file'
        else:
            purl.schema = 'pwd'

    log.debug('  -> %s', purl)

    if purl.schema not in list(context.keys()):

        ret = purl
        log.debug('             = %s (%s)', ret, list(context.keys()))

    else:

        expand = True

        # we expect hostname elements to be absent for schemas we expand
        if purl.host:
            raise ValueError('URLs cannot specify `host` for expanded schemas')

        if purl.schema == 'file':
            # we leave `file://` URLs unaltered
            ret = purl
            expand = False

        elif purl.schema == 'pwd' and 'pwd' not in context:
            ret = ru.Url(os.getcwd())

        else:
            ret = ru.Url(context[purl.schema])

        if expand:
            ret.path += '/%s' % purl.path

        if expand:
            log.debug('   expand with %s', context.get(purl.schema))

    log.debug('             > %s', ret)

    return ret


# ------------------------------------------------------------------------------




__copyright__ = 'Copyright 2016-2023, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import math

import radical.utils as ru

from ...   import constants as rpc
from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class JSRUN(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        self._erf    : bool = False
        self._command: str  = ''

        LaunchMethod.__init__(self, name, lm_cfg, rm_info, log, prof)


    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):

        lm_info = {'env'    : env,
                   'env_sh' : env_sh,
                   'command': ru.which('jsrun'),
                   'erf'    : False}

        if '_erf' in self.name.lower():
            lm_info['erf'] = True

        return lm_info


    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        self._env     = lm_info['env']
        self._env_sh  = lm_info['env_sh']
        self._command = lm_info['command']

        assert self._command

        self._erf     = lm_info['erf']


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        pass


    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        if not task['description']['executable']:
            return False, 'no executable'

        return True, ''


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        return ['. $RP_PILOT_SANDBOX/%s' % self._env_sh]


    # --------------------------------------------------------------------------
    #
    def _create_resource_set_file(self, slots, uid, sandbox):
        '''
        This method takes as input a Task slots and creates the necessary
        resource set file (Explicit Resource File, ERF). This resource set
        file is then used by jsrun to place and execute tasks on nodes.

        An example of a resource file is:

        * Task 1: 2 MPI procs, 2 threads per process and 2 gpus per process*

            rank 0 : { host: 1; cpu:  {0, 1}; gpu: {0,1} }
            rank 1 : { host: 1; cpu: {22,23}; gpu: {3,4} }

        * Task 2: 2 MPI procs, 1 thread per process and 1 gpus per process*

            rank 0 : { host: 2; cpu:  {7}; gpu: {2} }
            rank 1 : { host: 2; cpu: {30}; gpu: {5} }

        Parameters
        ----------
        slots : List of dictionaries.

            The slots that the task will be placed. A slot has the following
            format:

            {"ranks"         : [{"node_name" : "a",
                                 "node_id"   : "1",
                                 "core_map"  : [[0, 1]],
                                 "gpu_map"   : [[0]],
                                 "lfs"       : 0,
                                 "mem"       : 0
                                }]
            }

        uid     : task ID (string)
        sandbox : task sandbox (string)
        '''

        # https://github.com/olcf-tutorials/ERF-CPU-Indexing
        # `cpu_index_using: physical` causes the following issue
        # "error in ptssup_mkcltsock_afunix()"
        rs_str  = 'cpu_index_using: logical\n'

        base_id = 0
        for slot_ranks in slots['ranks']:

            ranks_per_rs  = len(slot_ranks['core_map'])
            rank_ids      = [str(r + base_id) for r in range(ranks_per_rs)]
            base_id      += ranks_per_rs

            core_id_sets = []
            for core_map in slot_ranks['core_map']:
                core_ids = [str(cid) for cid in core_map]
                core_id_sets.append('{%s}' % ','.join(core_ids))

            rs_str += 'rank: %s : {'    % ','.join(rank_ids)
            rs_str += ' host: %s;'      % str(slot_ranks['node_id'])
            rs_str += ' cpu: %s'        % ','.join(core_id_sets)
            if slot_ranks['gpu_map']:
                slot_gpus = slot_ranks['gpu_map'][0]
                assert slot_ranks['gpu_map'].count(slot_gpus) == ranks_per_rs
                rs_str += '; gpu: {%s}' % ','.join([str(g) for g in slot_gpus])
            rs_str += ' }\n'

        rs_name = '%s/%s.rs' % (sandbox, uid)
        with ru.ru_open(rs_name, 'w') as fout:
            fout.write(rs_str)

        return rs_name


    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        uid   = task['uid']
        td    = task['description']
        slots = task['slots']

        assert slots['ranks'], 'task.slots.ranks not defined'

        if self._erf:

            cmd_options = '--erf_input %s' % self._create_resource_set_file(
                slots, uid, task['task_sandbox_path'])

        else:

            # JSRun uses resource sets (RS) to configure a node representation
            # for a job/task: https://docs.olcf.ornl.gov/systems/\
            #                 summit_user_guide.html#resource-sets

            rs             = len(slots['ranks'])
            slot_ranks     = slots['ranks'][0]
            # physical cores per rank
            cores_per_rank = math.ceil(len(slot_ranks['core_map'][0]) /
                             self._rm_info['threads_per_core'])
            ranks_per_rs   = len(slot_ranks['core_map'])
            cores_per_rs   = cores_per_rank * ranks_per_rs

            gpus_per_rs  = 0
            if slot_ranks['gpu_map']:
                slot_gpus = slot_ranks['gpu_map'][0]
                assert slot_ranks['gpu_map'].count(slot_gpus) == ranks_per_rs
                gpus_per_rs = len(slot_gpus)

            # -n: number of RS
            # -a: number of MPI tasks (ranks)     per RS
            # -c: number of CPUs (physical cores) per RS
            # -g: number of GPUs                  per RS
            cmd_options = '-n%d -a%d -c%d -g%d' % (
                rs, ranks_per_rs, cores_per_rs, gpus_per_rs)

            if gpus_per_rs:
                # -r: number of RS per host (node)
                max_rs_per_node = self._rm_info['gpus_per_node'] // gpus_per_rs
                if rs > max_rs_per_node:
                    # number of nodes that specified must evenly divide the
                    # number of resource sets, otherwise an error is returned
                    max_rs_per_node = math.gcd(rs, max_rs_per_node)
                else:
                    max_rs_per_node = min(rs, max_rs_per_node)
                cmd_options += ' -r%d' % max_rs_per_node

            # -b: binding of tasks within a resource set (none, rs, or packed:N)
            if ranks_per_rs > 1:
                if td['threading_type'] == rpc.OpenMP:
                    # for OpenMP threads RP will set:
                    #    export OMP_NUM_THREADS=<threads_per_rank>
                    cmd_options += ' -b packed:%d' % cores_per_rank
            else:
                cmd_options += ' -b rs'

        if td['gpus_per_rank'] and td['gpu_type'] == rpc.CUDA:
            # from https://www.olcf.ornl.gov/ \
            #             wp-content/uploads/2018/11/multi-gpu-workshop.pdf
            #
            # CUDA with    MPI, use jsrun --smpiargs="-gpu"
            # CUDA without MPI, use jsrun --smpiargs="off"
            #
            # This is set for CUDA tasks only
            if td['ranks'] > 1:
                # MPI is enabled
                cmd_options += ' --smpiargs="-gpu"'
            else:
                cmd_options += ' --smpiargs="off"'

        cmd = '%s %s %s' % (self._command, cmd_options, exec_path)
        return cmd.rstrip()


    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        ret = 'test -z "$PMIX_RANK" || export RP_RANK=$PMIX_RANK\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def get_exec(self, task):

        td          = task['description']
        task_exec   = td['executable']
        task_args   = td.get('arguments')
        task_argstr = self._create_arg_string(task_args)
        command     = '%s %s' % (task_exec, task_argstr)

        return command.rstrip()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2014-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import time
import radical.utils  as ru

from ..  import utils as rpu


# ------------------------------------------------------------------------------
#
class Agent_n(rpu.Worker):

    # This is a sub-agent.  It does not do much apart from starting
    # agent components and watching them, which is all taken care of in the
    # `Worker` base class (or rather in the `Component` base class of `Worker`).

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._cfg      = cfg
        self._pid      = cfg.pid
        self._pmgr     = cfg.pmgr
        self._pwd      = cfg.pilot_sandbox
        self._sid      = cfg.sid
        self._reg_addr = cfg.reg_addr

        # log / profile via session until component manager is initialized
        self._session = session
        self._log     = session._log
        self._prof    = session._prof

        self._starttime   = time.time()
        self._final_cause = None

        # this is the earliest point to sync bootstrap and agent profiles
        self._prof.prof('hostname', uid=self._pid, msg=ru.get_hostname())
        self._prof.prof('sub_agent_start', uid=self._pid)

        # expose heartbeat channel to sub-agents, bridges and components,
        # and start those
        self._cmgr = rpu.ComponentManager(self._cfg)
        self._cfg.heartbeat = self._cmgr.cfg.heartbeat

        self._cmgr.start_bridges()
        self._cmgr.start_components()

        # at this point the session is up and connected, and it should have
        # brought up all communication bridges and components.  We are
        # ready to rumble!
        rpu.Worker.__init__(self, self._cfg, session)


    # --------------------------------------------------------------------------
    def finalize(self):

        self._prof.prof('sub_agent_stop', uid=self._pid, msg=self._uid)


# ------------------------------------------------------------------------------




#!/usr/bin/env python3


__copyright__ = "Copyright 2014-2015, http://radical.rutgers.edu"
__license__   = "MIT"

import sys
import radical.pilot as rp


# DESCRIPTION: Tutorial 1: A Simple Workload consisting of a Bag-of-Tasks
#                             submitted to multiple machines
#

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


# ------------------------------------------------------------------------------
#
def pilot_state_cb (pilot, state):

    if not pilot:
        return

    print("[Callback]: Pilot '%s' state: %s." % (pilot.uid, state))

    if state == rp.FAILED:
        sys.exit (1)


# ------------------------------------------------------------------------------
#
def task_state_cb (task, state):

    if not task:
        return

    print("[Callback]: task %s on %s: %s." % (task.uid, task.pilot_id, state))

    if state == rp.FAILED:
        print("stderr: %s" % task.stderr)
        sys.exit(2)


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # we can optionally pass session name to RP
    if len(sys.argv) > 1:
        session_name = sys.argv[1]
    else:
        session_name = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session(name=session_name)
    print("session id: %s" % session.uid)

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        #
        # Change the user name below if you are using a remote resource
        # and your username on that resource is different from the username
        # on your local machine.
        #
        c = rp.Context('ssh')
        c.user_id = "username"
      # c.user_pass = "PutYourPasswordHere"
        session.add_context(c)

        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        print("Initializing Pilot Manager ...")
        pmgr = rp.PilotManager(session=session)

        # Register our callback with the PilotManager. This callback will get
        # called every time any of the pilots managed by the PilotManager
        # change their state.
        pmgr.register_callback(pilot_state_cb)

        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        #
        # If you want to run this example on XSEDE/ACCESS Gordon and Bridges2, you have
        # to add your allocation ID by setting the project attribute for each
        # pilot description ot it.
        #
        # A list of preconfigured resources can be found at:
        # https://radicalpilot.readthedocs.io/en/stable/ \
        #        machconf.html#preconfigured-resources
        #

        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        # The pilot_list will contain the description of the pilot that will be
        # submitted
        pilot_list = list()

        # Create the description of the first pilot and add it to the list
        pdesc = rp.PilotDescription ()
        pdesc.resource = "access.gordon"
        pdesc.runtime  = 10
        pdesc.cores    = 1
        pdesc.cleanup  = True
        pdesc.project  = ''
        pilot_list.append(pdesc)

        # Create the description of the secind pilot and add it to the list
        pdesc2 = rp.PilotDescription ()
        pdesc2.resource = "access.bridges2"
        pdesc2.runtime  = 10
        pdesc2.cores    = 1
        pdesc2.cleanup  = True
        pdesc2.project  = ''
        pilot_list.append(pdesc2)

        # Continue adding pilot by creating a new descrption and appending it to
        # the list.

        # Submit the pilot list to the Pilot Manager. Actually all the pilots are
        # submitted to the Pilot Manager at once.
        print("Submitting  Pilots to Pilot Manager ...")
        pilots = pmgr.submit_pilots(pilot_list)

        # Combine the Pilot, the Tasks and a scheduler via
        # a TaskManager object. The scheduler that supports multi-pilot sessions
        # is Round Robin. Direct Submittion does not.
        print("Initializing Task Manager ...")
        tmgr = rp.TaskManager (session=session,
                               scheduler=rp.SCHEDULER_ROUND_ROBIN)

        # Register our callback with the TaskManager. This callback will get
        # called every time any of the tasks managed by the TaskManager
        # change their state.
        tmgr.register_callback(task_state_cb)

        # Add the created Pilot to the TaskManager.
        print("Registering  Pilots with Task Manager ...")
        tmgr.add_pilots(pilots)

        NUMBER_JOBS  = 64  # the total number of tasks to run

        # submit tasks to pilot job
        taskdesc_list = []
        for i in range(NUMBER_JOBS):

            # -------- BEGIN USER DEFINED Task DESCRIPTION --------- #
            taskdesc = rp.TaskDescription()
            taskdesc.environment = {'task_NO': i}
            taskdesc.executable  = "/bin/echo"
            taskdesc.arguments   = ['I am Task number $task_NO from $HOSTNAME']
            taskdesc.cores       = 1
            # -------- END USER DEFINED Task DESCRIPTION --------- #

            taskdesc_list.append(taskdesc)

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        print("Submit Tasks to Task Manager ...")
        task_set = tmgr.submit_tasks (taskdesc_list)

        print("Waiting for tasks to complete ...")
        tmgr.wait_tasks()
        print("All tasks completed successfully!")


    except Exception as e:
        # Something unexpected happened in the pilot code above
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print("closing session")
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


# ------------------------------------------------------------------------------




# pylint: disable=protected-access

__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import copy
import time

import radical.utils as ru

from . import PilotManager
from . import states    as rps
from . import constants as rpc

from .staging_directives import complete_url


# ------------------------------------------------------------------------------
#
class Pilot(object):
    """Represent a resource overlay on a local or remote resource.

    Note:
        A Pilot cannot be created directly. The factory method
        :func:`radical.pilot.PilotManager.submit_pilots` has to be
        used instead.

    Example::

          pm = radical.pilot.PilotManager(session=s)
          pd = radical.pilot.PilotDescription()

          pd.resource = "local.localhost"
          pd.cores    = 2
          pd.runtime  = 5 # minutes

          pilot = pm.submit_pilots(pd)

    """

    # --------------------------------------------------------------------------
    # In terms of implementation, a Pilot is not much more than a dict whose
    # content are dynamically updated to reflect the state progression through
    # the PMGR components.  As a Pilot is always created via a PMGR, it is
    # considered to *belong* to that PMGR, and all activities are actually
    # implemented by that PMGR.
    #
    # Note that this implies that we could create Pilots before submitting them
    # to a PMGR, w/o any problems. (FIXME?)
    # --------------------------------------------------------------------------


    # --------------------------------------------------------------------------
    #
    def __init__(self, pmgr: PilotManager, descr):

        # sanity checks on description
        if not descr.runtime:
            raise ValueError('pilot runtime must be defined')

        if descr.runtime <= 0:
            raise ValueError('pilot runtime must be positive')

        if not descr.resource:
            raise ValueError('pilot target resource must be defined')


        # initialize state
        self._descr      = descr.as_dict()
        self._pmgr       = pmgr
        self._session    = self._pmgr.session
        self._prof       = self._session._prof
        self._uid        = self._descr.get('uid')
        self._state      = rps.NEW
        self._log        = pmgr._log
        self._pilot_dict = dict()
        self._callbacks  = dict()
        self._cache      = dict()    # cache of SAGA dir handles
        self._cb_lock    = ru.RLock()
        self._tmgr       = None

        # pilot failures can trigger app termination
        self._exit_on_error = self._descr.get('exit_on_error')

        # ensure uid is unique
        if self._uid:
            if not self._pmgr.check_uid(self._uid):
                raise ValueError('uid %s is not unique' % self._uid)
        else:
            self._uid = ru.generate_id('pilot.%(item_counter)04d', ru.ID_CUSTOM,
                                       ns=self._session.uid)

        for m in rpc.PMGR_METRICS:
            self._callbacks[m] = dict()

        # we always invoke the default state cb
        self._callbacks[rpc.PILOT_STATE][self._default_state_cb.__name__] = {
                'cb'      : self._default_state_cb,
                'cb_data' : None}

        # `as_dict()` needs `pilot_dict` and other attributes.  Those should all
        # be available at this point (apart from the sandboxes), so we now
        # query for those sandboxes.
        self._pilot_jsurl      = ru.Url()
        self._pilot_jshop      = ru.Url()
        self._endpoint_fs      = ru.Url()
        self._resource_sandbox = ru.Url()
        self._session_sandbox  = ru.Url()
        self._pilot_sandbox    = ru.Url()
        self._client_sandbox   = ru.Url()

        pilot = self.as_dict()

        self._pilot_jsurl, self._pilot_jshop \
                               = self._session._get_jsurl           (pilot)
        self._endpoint_fs      = self._session._get_endpoint_fs     (pilot)
        self._resource_sandbox = self._session._get_resource_sandbox(pilot)
        self._session_sandbox  = self._session._get_session_sandbox (pilot)
        self._pilot_sandbox    = self._session._get_pilot_sandbox   (pilot)
        self._client_sandbox   = self._session._get_client_sandbox()

        # contexts for staging url expansion
        # NOTE: no task sandboxes defined!
        self._rem_ctx = {'pwd'     : self._pilot_sandbox,
                         'client'  : self._client_sandbox,
                         'pilot'   : self._pilot_sandbox,
                         'resource': self._resource_sandbox,
                         'session' : self._session_sandbox,
                         'endpoint': self._endpoint_fs}

        self._loc_ctx = {'pwd'     : self._client_sandbox,
                         'client'  : self._client_sandbox,
                         'pilot'   : self._pilot_sandbox,
                         'resource': self._resource_sandbox,
                         'session' : self._session_sandbox,
                         'endpoint': self._endpoint_fs}


        # we need to expand plaaceholders in the sandboxes
        # FIXME: this code is a duplication from the pilot launcher code
        expand = dict()
        for k,v in pilot['description'].items():
            if v is None:
                v = ''
            expand['pd.%s' % k] = v
            if isinstance(v, str):
                expand['pd.%s' % k.upper()] = v.upper()
                expand['pd.%s' % k.lower()] = v.lower()
            else:
                expand['pd.%s' % k.upper()] = v
                expand['pd.%s' % k.lower()] = v

        self._endpoint_fs     .path  = self._endpoint_fs     .path % expand
        self._resource_sandbox.path  = self._resource_sandbox.path % expand
        self._session_sandbox .path  = self._session_sandbox .path % expand
        self._pilot_sandbox   .path  = self._pilot_sandbox   .path % expand


    # --------------------------------------------------------------------------
    #
    def __repr__(self):

        return '<%s object, uid %s>' % (self.__class__.__qualname__, self._uid)


    # --------------------------------------------------------------------------
    #
    def __str__(self):

        return str([self.uid, self.resource, self.state])


    # --------------------------------------------------------------------------
    #
    def _default_state_cb(self, pilot, state=None):

        uid   = self.uid
        state = self.state

        self._log.info("[Callback]: pilot %s state: %s.", uid, state)

        if state == rps.FAILED and self._exit_on_error:
            self._log.error("[Callback]: pilot '%s' failed (exit)", uid)

            # There are different ways to tell main...
            ru.cancel_main_thread('int')
          # raise RuntimeError('pilot %s failed - fatal!' % self.uid)
          # os.kill(os.getpid())
          # sys.exit()


    # --------------------------------------------------------------------------
    #
    def _update(self, pilot_dict):
        """Trigger an update.

        This will update the facade object after state changes etc, and is
        invoked by whatever component receiving that updated information.

        Returns:
             booL: True if state changed, False otherwise.

        """

        self._log.debug('update %s', pilot_dict['uid'])

        if pilot_dict['uid'] != self.uid:
            self._log.error('invalid uid: %s / %s', pilot_dict['uid'], self.uid)

        assert pilot_dict['uid'] == self.uid, 'update called on wrong instance'

        # NOTE: this method relies on state updates to arrive in order and
        #       without gaps.
        current = self.state
        target  = pilot_dict['state']

        if target not in [rps.FAILED, rps.CANCELED]:

            # ensure valid state transition
            state_diff = rps._pilot_state_value(target) - \
                         rps._pilot_state_value(current)
            if state_diff != 1:
                raise RuntimeError('%s: invalid state transition %s -> %s',
                                   self.uid, current, target)

        self._state = target

        # keep all information around
        self._pilot_dict = copy.deepcopy(pilot_dict)

        # invoke pilot specific callbacks
        # FIXME: this iteration needs to be thread-locked!
        for _,cb_val in self._callbacks[rpc.PILOT_STATE].items():

            cb      = cb_val['cb']
            cb_data = cb_val['cb_data']

            self._log.debug('call %s', cb)

            self._log.debug('%s calls cb %s', self.uid, cb)

            if cb_data: cb([self], cb_data)
            else      : cb([self])

        # ask pmgr to invoke any global callbacks
        self._pmgr._call_pilot_callbacks(self)


    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        """Dictionary representation.

        Returns:
             dict: a Python dictionary representation of the object.

        """

        ret = {'session'          : self.session.uid,
               'pmgr'             : self.pmgr.uid,
               'uid'              : self.uid,
               'type'             : 'pilot',
               'state'            : self.state,
               'log'              : self.log,
               'stdout'           : self.stdout,
               'stderr'           : self.stderr,
               'resource'         : self.resource,
               'endpoint_fs'      : str(self._endpoint_fs),
               'resource_sandbox' : str(self._resource_sandbox),
               'session_sandbox'  : str(self._session_sandbox),
               'pilot_sandbox'    : str(self._pilot_sandbox),
               'client_sandbox'   : str(self._client_sandbox),
               'js_url'           : str(self._pilot_jsurl),
               'js_hop'           : str(self._pilot_jshop),
               'description'      : self.description,  # this is a deep copy
               'resource_details' : self.resource_details
              }

        return ret


    # --------------------------------------------------------------------------
    #
    @property
    def session(self):
        """Session: The pilot's session."""

        return self._session


    # --------------------------------------------------------------------------
    #
    @property
    def pmgr(self):
        """PilotManager: The pilot's manager."""

        return self._pmgr


    # -------------------------------------------------------------------------
    #
    @property
    def resource_details(self):
        """dict: agent level resource information."""

        return self._pilot_dict.get('resource_details')


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        """str: The pilot's unique identifier within a :class:`PilotManager`."""

        return self._uid


    # --------------------------------------------------------------------------
    #
    @property
    def state(self):
        """str: The current :py:mod:`state <radical.pilot.states>` of the pilot."""

        return self._state


    # --------------------------------------------------------------------------
    #
    @property
    def log(self):
        """list[tuple]: A list of human readable [timestamp, string] tuples describing
        various events during the pilot's lifetime.  Those strings are not
        normative, only informative!
        """

        return self._pilot_dict.get('log')


    # --------------------------------------------------------------------------
    #
    @property
    def stdout(self):
        """str: A snapshot of the pilot's STDOUT stream.

        If this property is queried before the pilot has reached
        'DONE' or 'FAILED' state it will return None.

        Warning:
            This can be inefficient. Output may be incomplete and/or filtered.

        """

        return self._pilot_dict.get('stdout')


    # --------------------------------------------------------------------------
    #
    @property
    def stderr(self):
        """str: A snapshot of the pilot's STDERR stream.

        If this property is queried before the pilot has reached
        'DONE' or 'FAILED' state it will return None.

        Warning:
            This can be inefficient.  Output may be incomplete and/or filtered.

        """

        return self._pilot_dict.get('stderr')


    # --------------------------------------------------------------------------
    #
    @property
    def resource(self):
        """str: The resource tag of this pilot."""

        return self._descr.get('resource')


    # --------------------------------------------------------------------------
    #
    @property
    def pilot_sandbox(self):
        """str: The full sandbox URL of this pilot, if that is already
        known, or 'None' otherwise.
        """

        # NOTE: The pilot has a sandbox property, containing the full sandbox
        #       path, which is used by the pmgr to stage data back and forth.
        #       However, the full path as visible from the pmgr side might not
        #       be what the agent is seeing, specifically in the case of
        #       non-shared filesystems (OSG).  The agent thus uses
        #       `$PWD` as sandbox, with the assumption that this will
        #       get mapped to whatever is here returned as sandbox URL.
        #
        #       There is thus implicit knowledge shared between the RP client
        #       and the RP agent that `$PWD` *is* the sandbox!  The same
        #       implicitly also holds for the staging area, which is relative
        #       to the pilot sandbox.
        if self._pilot_sandbox:
            return str(self._pilot_sandbox)


    @property
    def endpoint_fs(self):
        """radical.utils.Url: The URL which is internally used to access the
        target resource's root file system.
        """
        return self._endpoint_fs

    @property
    def resource_sandbox(self):
        """radical.utils.Url: The full URL of the path that RP considers the
        resource sandbox, i.e., the sandbox on the target resource's file system
        which is shared by all sessions which access that resource.
        """
        return self._resource_sandbox


    @property
    def session_sandbox(self):
        """radical.utils.Url: The full URL of the path that RP considers the
        session sandbox on the target resource's file system which is shared by
        all pilots which access that resource in the current session.
        """
        return self._session_sandbox

    @property
    def client_sandbox(self):
        return self._client_sandbox


    # --------------------------------------------------------------------------
    #
    @property
    def description(self):
        """dict: The description the pilot was started with, as a dictionary."""

        return copy.deepcopy(self._descr)


    # --------------------------------------------------------------------------
    #
    def register_callback(self, cb, metric=rpc.PILOT_STATE, cb_data=None):
        """Add callback for state changes.

        Registers a callback function that is triggered every time the
        pilot's state changes.

        All callback functions need to have the same signature::

            def cb(obj, state)

        where ``obj`` is a handle to the object that triggered the callback
        and ``state`` is the new state of that object.  If *cb_data* is given,
        then the *cb* signature changes to
        ::

            def cb(obj, state, cb_data)

        and *cb_data* are passed along.

        """

        if metric not in rpc.PMGR_METRICS :
            raise ValueError ("invalid pmgr metric '%s'" % metric)

        with self._cb_lock:
            cb_id = id(cb)
            self._callbacks[metric][cb_id] = {'cb'      : cb,
                                              'cb_data' : cb_data}


    # --------------------------------------------------------------------------
    #
    def unregister_callback(self, cb, metric=rpc.PILOT_STATE):

        if metric and metric not in rpc.PMGR_METRICS :
            raise ValueError ("invalid pmgr metric '%s'" % metric)

        if not metric:
            metrics = rpc.PMGR_METRICS
        elif isinstance(metric, list):
            metrics =  metric
        else:
            metrics = [metric]

        with self._cb_lock:

            for metric in metrics:

                if cb:
                    to_delete = [id(cb)]
                else:
                    to_delete = list(self._callbacks[metric].keys())

                for cb_id in to_delete:

                    if cb_id not in self._callbacks[metric]:
                        raise ValueError("unknown callback '%s'" % cb_id)

                    del self._callbacks[metric][cb_id]


    # --------------------------------------------------------------------------
    #
    def wait(self, state=None, timeout=None):
        """Block for state change.

        Returns when the pilot reaches a specific state or
        when an optional timeout is reached.

        Arguments:
            state (list[str]):
                The :py:mod:`state(s) <radical.pilot.states>` that pilot has to reach in
                order for the call to return.

                By default `wait` waits for the pilot to reach a **final**
                state, which can be one of the following:

                * :data:`radical.pilot.states.DONE`
                * :data:`radical.pilot.states.FAILED`
                * :data:`radical.pilot.states.CANCELED`
            timeout (float):
                Optional timeout in seconds before the call returns regardless
                whether the pilot has reached the desired state or not.  The
                default value **None** never times out.

        """

        if   not state                  : states = rps.FINAL
        elif not isinstance(state, list): states = [state]
        else                            : states = state


        if self.state in rps.FINAL:
            # we will never see another state progression.  Raise an error
            # (unless we waited for this)
            if self.state in states:
                return

            # FIXME: do we want a raise here, really?  This introduces a race,
            #        really, on application level
            # raise RuntimeError("can't wait on a pilot in final state")
            return self.state

        start_wait = time.time()
        while self.state not in states:

            time.sleep(0.1)
            if timeout and (timeout <= (time.time() - start_wait)):
                break

            if self._pmgr._terminate.is_set():
                break

        return self.state


    # --------------------------------------------------------------------------
    #
    def cancel(self):
        """Cancel the pilot."""

        # clean connection cache
        try:
            for key in self._cache:
                self._cache[key].close()
            self._cache = dict()

        except:
            pass

        self._pmgr.cancel_pilots(self.uid)


    # --------------------------------------------------------------------------
    #
    def attach_tmgr(self, tmgr) -> None:

        if self._tmgr:
            raise RuntimeError('this pilot is already attached to %s'
                               % self._tmgr.uid)
        self._tmgr = tmgr

      # if self._task_waitpool:
      #     self._tmgr.submit_tasks(self._task_waitpool)
      #
      # if self._raptor_waitpool:
      #     self._tmgr.submit_tasks(self._raptor_waitpool)


    # --------------------------------------------------------------------------
    #
    def submit_tasks(self, descriptions):

        for descr in descriptions:
            descr.pilot = self.uid

        if not self._tmgr:

          # self._task_waitpool.append(descriptions)
          # return  # FIXME: cannot return tasks here

            raise RuntimeError('pilot is not attached to a task manager, yet')

        return self._tmgr.submit_tasks(descriptions)


    # --------------------------------------------------------------------------
    #
    def submit_raptors(self, descriptions):

        if not self._tmgr:

          # self._task_waitpool.append(descriptions)
          # return  # FIXME: cannot return tasks here

            raise RuntimeError('pilot is not attached to a task manager, yet')

        return self._tmgr.submit_raptors(descriptions, self.uid)


    # --------------------------------------------------------------------------
    #
    def prepare_env(self, env_name, env_spec):
        """Prepare a virtual environment.

        Request the preparation of a task or worker environment on the target
        resource.  This call will block until the env is created.

        Arguments:
            env_name (str): name of the environment to prepare.
            env_spec (dict): specification of the environment to prepare,
                like::

                    {'type'    : 'venv',
                     'version' : '3.7',
                     'pre_exec': ['module load python'],
                     'setup'   : ['radical.pilot==1.0', 'pandas']},

                    {'type'    : 'conda',
                     'version' : '3.8',
                     'setup'   : ['numpy']}

                    {'type'   : 'conda',
                     'version': '3.8',
                     'path'   : '/path/to/ve',
                     'setup'  : ['numpy']}

                where the *type* specifies the environment type, *version* specifies the
                Python version to deploy, and *setup* specifies how the environment is
                to be prepared.  If *path* is specified the env will be created at that
                path.  If *path* is not specified, RP will place the named env in the
                pilot sandbox (under :file:`env/named_env_{name}`). If a VE exists at that
                path, it will be used as is (an update is not performed). *pre_exec*
                commands are executed before env creation and setup are attempted.

        Note:
            The `version` specifier is only interpreted up to minor version;
            subminor and less are ignored.

        """

        self.rpc('prepare_env', {'env_name': env_name,
                                 'env_spec': env_spec})


    # --------------------------------------------------------------------------
    #
    def rpc(self, rpc, args):
        """Remote procedure call.

        Send a pilot command, wait for the response, and return the result.
        This is basically an RPC into the pilot.
        """

        reply = self._session._dbs.pilot_rpc(self.uid, self.uid, rpc, args)

        return reply


    # --------------------------------------------------------------------------
    #
    def stage_in(self, sds):
        """Stage files "in".

        Stages the content of the :py:mod:`~radical.pilot.staging_directives`
        to the pilot sandbox.

        Please note the documentation of
        :func:`radical.pilot.staging_directives.complete_url` for details on
        path and sandbox semantics.
        """

        sds = ru.as_list(sds)

        for sd in sds:
            sd['prof_id'] = self.uid
            sd['source'] = str(complete_url(sd['source'], self._loc_ctx, self._log))
            sd['target'] = str(complete_url(sd['target'], self._rem_ctx, self._log))

        # ask the pmgr to send the staging requests to the stager
        self._pmgr._pilot_staging_input(sds)

        return [sd['target'] for sd in sds]


    # --------------------------------------------------------------------------
    #
    def stage_out(self, sds=None):
        """Stage data "out".

        Fetches the content of the :py:mod:`~radical.pilot.staging_directives`
        from the pilot sandbox.

        Please note the documentation of
        :func:`radical.pilot.staging_directives.complete_url` for details on
        path and sandbox semantics.
        """

        sds = ru.as_list(sds)

        if not sds:
            sds = [{'source': 'pilot:///staging_output.tgz',
                    'target': 'client:///staging_output.tgz',
                    'action': rpc.TRANSFER}]

        for sd in sds:
            sd['prof_id'] = self.uid

        for sd in sds:
            sd['source'] = str(complete_url(sd['source'], self._rem_ctx, self._log))
            sd['target'] = str(complete_url(sd['target'], self._loc_ctx, self._log))

        # ask the pmgr to send the staging reuests to the stager
        self._pmgr._pilot_staging_output(sds)

        return [sd['target'] for sd in sds]


# ------------------------------------------------------------------------------



#!/usr/bin/env python3

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os

os.environ['RADICAL_PILOT_BULK_CB'] = 'True'

import random

import radical.pilot as rp

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


# ------------------------------------------------------------------------------
#
def pilot_state_cb(pilots):
    """
    this callback is invoked on all pilot state changes
    """

    # Callbacks happen in a different thread than the main application thread,
    # i.e., they are asynchronous.  However, this means that a 'sys.exit()'
    # does not end the application but just a thread (in this case the
    # pilot_manager_controller thread).  Thus, we wrap all threads in their own
    # try/except clauses, and then translate the `sys.exit()` into a
    # 'thread.interrupt_main()' call. This raises a 'KeyboardInterrupt' in the
    # main thread which can be interpreted by your application, for example to
    # initiate a clean shutdown via `session.close()` (see code below). The
    # same `KeyboardShutdown` is also raised when you interrupt the application
    # via `^C`.
    #
    # Note that other error handling semantics are available, depending on your
    # application requirements. For example, upon a pilot failure, the
    # application could spawn a replacement pilot, or reduce the number of
    # tasks to match the remaining set of pilots.

    for pilot in pilots:
        print("[Callback]: Pilot '%s' state: %s." % (pilot.uid, pilot.state))

        if pilot.state == rp.FAILED:
            print(pilot.log[-1])  # Get the last log message


# -----------------------------------------------------------------------------
#
def task_state_cb(tasks):
    """ this callback is invoked on all task state changes """

    # The approach to task state callbacks is the same as the one to
    # pilot state callbacks. Only difference is that task state
    # callbacks are invoked by the task manager on changes of task
    # states.
    #
    # The example below does not really create any Task object, we only
    # include the callback here for documentation on the approaches to error
    # handling.
    #
    # Note that other error handling semantics are available, depending on your
    # application requirements. For example, upon tasks failure, the
    # application could spawn replacement tasks, or spawn a pilot on a
    # different resource which might be better equipped to handle the task
    # payload.

    for task in tasks:
        print('  task %s: %s' % (task.uid, task.state))

        if task.state == rp.FAILED:
            print('                  : %s' % task.stderr)


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # This example shows how simple error handling can be asynchronously
    # implemented using callbacks (see above).

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyway...
    session = rp.Session()
    print("session id: %s" % session.uid)

    # all other RP code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus
    # tear the whole RP stack down via a 'session.close()' call in the
    # 'finally' clause...
    try:

        # create the task and pilot managers.
        tmgr = rp.TaskManager(session=session)
        pmgr = rp.PilotManager(session=session)

        # Register our callbacks with the managers. The callbacks will get
        # called every time any of the pilots or tasks change their state,
        # including on FAILED states.
        tmgr.register_callback(task_state_cb)
        pmgr.register_callback(pilot_state_cb)

        # Create a local pilot.
        pd = rp.PilotDescription()
        pd.resource  = "local.localhost"
        pd.cores     = 64
        pd.runtime   = 60

        pilot = pmgr.submit_pilots(pd)
        tmgr.add_pilots(pilot)


        # we submit n tasks, some of which will fail.
        n    = 1024 * 3
        tds = list()
        for _ in range(n):
            td = rp.TaskDescription()
            if random.random() < 0.5:
                td.executable = '/bin/true'
            else:
                td.executable = '/bin/fail'
            tds.append(td)

        # submit the tasks...
        cus = tmgr.submit_tasks(tds)

        # ... and wait for their completion.
        state = tmgr.wait_tasks(state=rp.FINAL)


    except Exception as e:
        # Something unexpected happened in the pilot code above.
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also
        # catch SystemExit, which gets raised if the main thread exits for
        # some other reason.
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print("closing session")
        session.close()


# ------------------------------------------------------------------------------

